# encoding: utf-8

"""Provides abstracted cube-measure objects used as the basis for second-order measures.

There are several cube-measures that can appear in a cube-response, including
unweighted-counts, weighted-counts (aka. counts), means, and others.
"""

from __future__ import division

import numpy as np
from scipy.stats.contingency import expected_freq

from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.util import lazyproperty


class CubeMeasures(object):
    """Provides access to all cube-measure objects for this cube-result."""

    def __init__(self, cube, dimensions, slice_idx):
        self._cube = cube
        self._dimensions = dimensions
        self._slice_idx = slice_idx

    @lazyproperty
    def cube_means(self):
        """_BaseCubeMeans subclass object for this cube-result."""
        return _BaseCubeMeans.factory(self._cube, self._dimensions, self._slice_idx)

    @lazyproperty
    def cube_overlaps(self):
        """_BaseCubeOverlaps subclass object for this cube-result."""
        return _BaseCubeOverlaps.factory(self._cube, self._dimensions, self._slice_idx)

    @lazyproperty
    def cube_sum(self):
        """_BaseCubeSums subclass object for this cube-result."""
        return _BaseCubeSums.factory(self._cube, self._dimensions, self._slice_idx)

    @lazyproperty
    def cube_stddev(self):
        """_BaseCubeStdDev subclass object for this cube-result."""
        return _BaseCubeStdDev.factory(self._cube, self._dimensions, self._slice_idx)

    @lazyproperty
    def unweighted_cube_counts(self):
        """_BaseUnweightedCubeCounts subclass object for this cube-result."""
        return _BaseUnweightedCubeCounts.factory(
            self._cube, self._dimensions, self._slice_idx
        )

    @lazyproperty
    def weighted_cube_counts(self):
        """_BaseWeightedCounts subclass object for this cube-result."""
        return _BaseWeightedCubeCounts.factory(
            self._cube, self._dimensions, self._slice_idx
        )


class _BaseCubeMeasure(object):
    """Base class for all cube-measure objects."""

    def __init__(self, dimensions):
        self._dimensions = dimensions

    @classmethod
    def _slice_idx_expr(cls, cube, slice_idx):
        """Return np.s_ advanced-indexing slice object to extract values for slice_idx.

        `cube` can contain data for multiple `_Slice` objects. The returned `numpy`
        advanced-indexing expression selects only those values of `cube` that pertain
        to the slice indicated by `slice_idx`.
        """
        # --- for a 2D cube we take the whole thing (1D is not expected here) ---
        if cube.ndim < 3:
            return np.s_[:]

        # --- if 0th dimension of a >2D cube is MR, we only take the "Selected" portion
        # --- of the indicated initial-MR subvar, because the slice is to represent the
        # --- values for "respondents who selected" that MR response (and not those who
        # --- didn't select it or did not respond).
        if cube.dimension_types[0] == DT.MR:
            return np.s_[slice_idx, 0]

        # --- for other 3D cubes we just select the 2D "table" portion associated with
        # --- the `slice_idx`-th table dimension element.
        return np.s_[slice_idx]


# === MEANS ===


class _BaseCubeMeans(_BaseCubeMeasure):
    """Base class for mean cube-measure variants."""

    def __init__(self, dimensions, means):
        super(_BaseCubeMeans, self).__init__(dimensions)
        self._means = means

    @classmethod
    def factory(cls, cube, dimensions, slice_idx):
        """Return _BaseCubeMeans subclass instance appropriate to `cube`.

        Raises `ValueError` if the cube-result does not include a cube-means measure.
        """
        if cube.means is None:
            raise ValueError("cube-result does not contain cube-means measure")
        dimension_types = cube.dimension_types[-2:]
        CubeMeansCls = (
            _MrXMrCubeMeans
            if dimension_types == (DT.MR, DT.MR)
            else _MrXCatCubeMeans
            if dimension_types[0] == DT.MR
            else _CatXMrCubeMeans
            if dimension_types[1] == DT.MR
            else _CatXCatCubeMeans
        )
        return CubeMeansCls(
            dimensions, cube.means[cls._slice_idx_expr(cube, slice_idx)]
        )

    @lazyproperty
    def means(self):
        """2D np.float64 ndarray of cube means."""
        raise NotImplementedError(  # pragma: no cover
            "`%s` must implement `.means`" % type(self).__name__
        )


class _CatXCatCubeMeans(_BaseCubeMeans):
    """Means cube-measure for a slice with no MR dimensions."""

    @lazyproperty
    def means(self):
        """2D np.float64 ndarray of means for each valid matrix cell."""
        return self._means


class _CatXMrCubeMeans(_BaseCubeMeans):
    """Means cube-measure for a NOT_MR_X_MR slice.

    Note that the rows-dimensions need not actually be CAT.
    """

    @lazyproperty
    def means(self):
        """2D np.float64 ndarray of means for each valid matrix cell."""
        return self._means[:, :, 0]


class _MrXCatCubeMeans(_BaseCubeMeans):
    """Means cube-measure for an MR_X_NOT_MR slice.

    Note that the columns-dimension need not actually be CAT.
    """

    @lazyproperty
    def means(self):
        """2D np.float64 ndarray of means for each valid matrix cell."""
        return self._means[:, 0, :]


class _MrXMrCubeMeans(_BaseCubeMeans):
    """Means cube-measure for an MR_X_MR slice."""

    @lazyproperty
    def means(self):
        """2D np.float64 ndarray of means for each valid matrix cell."""
        # --- indexing is: all-rows, sel-only, all-cols, sel-only ---
        return self._means[:, 0, :, 0]


# === OVERLAPS ===


class _BaseCubeOverlaps(_BaseCubeMeasure):
    """Base class for overlap cube-measure variants."""

    def __init__(self, dimensions, overlaps, valid_overlaps):
        super(_BaseCubeOverlaps, self).__init__(dimensions)
        self._overlaps = overlaps
        self._valid_overlaps = valid_overlaps

    @classmethod
    def factory(cls, cube, dimensions, slice_idx):
        """Return _BaseCubeOverlaps subclass instance appropriate to `cube`.

        Raises `ValueError` if the cube-result does not include a cube-overlaps measure
        or if it doesn't include valid-cube-overlaps measure.
        """
        if cube.overlaps is None:
            raise ValueError(
                "cube-result does not contain cube-overlaps measure"
            )  # pragma: no cover
        if cube.valid_overlaps is None:
            raise ValueError(
                "cube-result does not contain cube-valid-overlaps measure"
            )  # pragma: no cover

        dimension_types = tuple(d.dimension_type for d in dimensions)
        idx_expr = cls._slice_idx_expr(cube, slice_idx)
        args = (dimensions, cube.overlaps[idx_expr], cube.valid_overlaps[idx_expr])
        return (
            _MrXMrOverlaps(*args)
            if dimension_types == (DT.MR, DT.MR)
            else _CatXMrOverlaps(*args)
        )

    @lazyproperty
    def overlaps(self):
        """3D np.float64 ndarray of cube overlaps."""
        raise NotImplementedError(  # pragma: no cover
            "`%s` must implement `.overlaps`" % type(self).__name__
        )


class _CatXMrOverlaps(_BaseCubeOverlaps):
    """Overlaps cube-measure for a NOT_MR_X_MR slice."""

    @lazyproperty
    def tile_repetitions(self):
        """tuple of number of repetitions of selected base matrix."""
        return (self._overlaps.shape[0], 1, 1)

    @lazyproperty
    def selected_bases(self):
        """3D np.float64 ndarray of selected overlaps between MR subvariables, per cat.

        For a CAT x MR matrix, the overlaps are calculated for each category, and then
        for each subvariables pair for that category (which will produce a square
        matrix for each category). So the output shape that we get back from the
        database is CAT x MR_SUBVAR x MR_SEL x MR_SUBVAR (the last one being the result
        of the `cube_overlap` measure, and representing the "pairing" with each subvar
        of the previous MR_SUBVAR dimension.

        From this shape, we only need the "Selected" part of the MR_SEL dimension, so
        we need to select the 0th element along the 2nd axis [:, :, 0]. But we also need
        to add all of the categories together, since that's what's used for the base
        value in all of the significance calculations. We then tile these bases, since
        they're the same for each row (all the categories added together). The tiling is
        done so that the API of this class can be uniform for all users.
        """
        return np.tile(np.sum(self._overlaps[:, :, 0], axis=0), self.tile_repetitions)

    @lazyproperty
    def valid_bases(self):
        """3D np.float64 ndarray of valid overlaps between MR subvariables, per cat.

        For a CAT x MR matrix, the overlaps are calculated for each category, and then
        for each subvariables pair for that category (which will produce a square
        matrix for each category). So the output shape that we get back from the
        database is CAT x MR_SUBVAR x MR_SEL x MR_SUBVAR (the last one being the result
        of the `cube_overlap` measure, and representing the "pairing" with each subvar
        of the previous MR_SUBVAR dimension.

        From this shape, we only need the "Selected" + "Other" part of the MR_SEL
        dimension (i.e. all except missing), so we need to add the 0th and the 1st
        element along the 2nd axis sum([:, :, 0:2]).
        """
        valids = np.sum(self._valid_overlaps[:, :, 0:2], axis=2)
        return np.tile(np.sum(valids, axis=0), self.tile_repetitions)


class _MrXMrOverlaps(_BaseCubeOverlaps):
    """Overlaps cube-measure for a MR_X_MR slice."""

    @lazyproperty
    def selected_bases(self):
        """3D np.float64 ndarray of selected overlaps bases between MR subvariables.

        For a MR x MR slice, the overlaps are calculated for each MR subvar row, and
        then for each subvariables pair for that row (which will produce a square
        matrix for each category). So the output shape that we get back from the
        database is MR_SUBVAR x MR_SEL x MR_SUBVAR x MR_SEL x MR_SUBVAR (the last one
        being the result of the `cube_overlap` measure, and representing the "pairing"
        with each subvar of the last MR_SUBVAR dimension.

        From this shape, we only need the "Selected" part of the last MR_SEL dimension,
        but we need both "Selected" and "Other" from the first MR_SEL dimension. We
        therefore need to select the 0th element along the 3rd axis, but we need to sum
        the 0th and 1st elements along the 1st axis (first MR_SEL dimension). The sum
        of the selected and other counts of the first MR_SEL is what's always used to
        represent bases of such crosstabs (we cannot simply ignore the other counts).
        """
        return np.sum(self._overlaps[:, 0:2, :, 0], axis=1)

    @lazyproperty
    def valid_bases(self):
        """3D np.float64 ndarray of valid overlaps between MR subvariables, per row.

        For a MR x MR slice, the overlaps are calculated for each row, and then
        for each subvariables pair for that row (which will produce a square matrix for
        each row). So the output shape that we get back from the database is
        MR_SUBVAR x MR_SEL x MR_SUBVAR x MR_SEL x MR_SUBVAR (the last one being the
        result of the `cube_overlap` measure, and representing the "pairing" with each
        subvar of the last MR_SUBVAR dimension.

        From this shape, we only need the "Selected" + "Other" part of both MR_SEL
        dimensions (i.e. all except missing), so we need to add the 0th and the 1st
        element along the 1st and 3rd axes sum([:, 0:2, :, 0:2], axis=(1, 3)).
        """
        return np.sum(self._valid_overlaps[:, 0:2, :, 0:2], axis=(1, 3))


# === STD DEV ===


class _BaseCubeStdDev(_BaseCubeMeasure):
    """Base class for stddev cube-measure variants."""

    def __init__(self, dimensions, stddev):
        super(_BaseCubeStdDev, self).__init__(dimensions)
        self._stddev = stddev

    @classmethod
    def factory(cls, cube, dimensions, slice_idx):
        """Return _BaseCubeStdDev subclass instance appropriate to `cube`.

        Raises `ValueError` if the cube-result does not include a cube-stddev measure.
        """
        if cube.stddev is None:
            raise ValueError("cube-result does not contain cube-stddev measure")
        dimension_types = cube.dimension_types[-2:]
        CubeSumsCls = (
            _MrXMrCubeStdDev
            if dimension_types == (DT.MR, DT.MR)
            else _MrXCatCubeStdDev
            if dimension_types[0] == DT.MR
            else _CatXMrCubeStdDev
            if dimension_types[1] == DT.MR
            else _CatXCatCubeStdDev
        )
        return CubeSumsCls(
            dimensions, cube.stddev[cls._slice_idx_expr(cube, slice_idx)]
        )

    @lazyproperty
    def stddev(self):
        """2D np.float64 ndarray of cube stddev."""
        raise NotImplementedError(  # pragma: no cover
            "`%s` must implement `.stddev`" % type(self).__name__
        )


class _CatXCatCubeStdDev(_BaseCubeStdDev):
    """StdDev cube-measure for a slice with no MR dimensions."""

    @lazyproperty
    def stddev(self):
        """2D np.float64 ndarray of stddev for each valid matrix cell."""
        return self._stddev


class _CatXMrCubeStdDev(_BaseCubeStdDev):
    """StdDev cube-measure for a NOT_MR_X_MR slice.

    Note that the rows-dimensions need not actually be CAT.
    """

    @lazyproperty
    def stddev(self):
        """2D np.float64 ndarray of stddev for each valid matrix cell."""
        return self._stddev[:, :, 0]


class _MrXCatCubeStdDev(_BaseCubeStdDev):
    """StdDev cube-measure for an MR_X_NOT_MR slice.

    Note that the columns-dimension need not actually be CAT.
    """

    @lazyproperty
    def stddev(self):
        """2D np.float64 ndarray of stddev for each valid matrix cell."""
        return self._stddev[:, 0, :]


class _MrXMrCubeStdDev(_BaseCubeStdDev):
    """StdDev cube-measure for an MR_X_MR slice."""

    @lazyproperty
    def stddev(self):
        """2D np.float64 ndarray of stddev for each valid matrix cell."""
        # --- indexing is: all-rows, sel-only, all-cols, sel-only ---
        return self._stddev[:, 0, :, 0]


# === SUMS ===


class _BaseCubeSums(_BaseCubeMeasure):
    """Base class for sum cube-measure variants."""

    def __init__(self, dimensions, sums):
        super(_BaseCubeSums, self).__init__(dimensions)
        self._sums = sums

    @classmethod
    def factory(cls, cube, dimensions, slice_idx):
        """Return _BaseCubeSums subclass instance appropriate to `cube`.

        Raises `ValueError` if the cube-result does not include a cube-sum measure.
        """
        if cube.sums is None:
            raise ValueError("cube-result does not contain cube-sum measure")
        dimension_types = cube.dimension_types[-2:]
        CubeSumsCls = (
            _MrXMrCubeSums
            if dimension_types == (DT.MR, DT.MR)
            else _MrXCatCubeSums
            if dimension_types[0] == DT.MR
            else _CatXMrCubeSums
            if dimension_types[1] == DT.MR
            else _CatXCatCubeSums
        )
        return CubeSumsCls(dimensions, cube.sums[cls._slice_idx_expr(cube, slice_idx)])

    @lazyproperty
    def sums(self):
        """2D np.float64 ndarray of cube sum."""
        raise NotImplementedError(  # pragma: no cover
            "`%s` must implement `.sum`" % type(self).__name__
        )


class _CatXCatCubeSums(_BaseCubeSums):
    """Sums cube-measure for a slice with no MR dimensions."""

    @lazyproperty
    def sums(self):
        """2D np.float64 ndarray of sum for each valid matrix cell."""
        return self._sums


class _CatXMrCubeSums(_BaseCubeSums):
    """Sums cube-measure for a NOT_MR_X_MR slice.

    Note that the rows-dimensions need not actually be CAT.
    """

    @lazyproperty
    def sums(self):
        """2D np.float64 ndarray of sum for each valid matrix cell."""
        return self._sums[:, :, 0]


class _MrXCatCubeSums(_BaseCubeSums):
    """Sums cube-measure for an MR_X_NOT_MR slice.

    Note that the columns-dimension need not actually be CAT.
    """

    @lazyproperty
    def sums(self):
        """2D np.float64 ndarray of sum for each valid matrix cell."""
        return self._sums[:, 0, :]


class _MrXMrCubeSums(_BaseCubeSums):
    """Sums cube-measure for an MR_X_MR slice."""

    @lazyproperty
    def sums(self):
        """2D np.float64 ndarray of sum for each valid matrix cell."""
        # --- indexing is: all-rows, sel-only, all-cols, sel-only ---
        return self._sums[:, 0, :, 0]


# === UNWEIGHTED COUNTS ===


class _BaseUnweightedCubeCounts(_BaseCubeMeasure):
    """Base class for unweighted-count cube-measure variants."""

    def __init__(self, dimensions, unweighted_counts, diff_nans):
        super(_BaseUnweightedCubeCounts, self).__init__(dimensions)
        self._unweighted_counts = unweighted_counts
        self._diff_nans = diff_nans

    @classmethod
    def factory(cls, cube, dimensions, slice_idx):
        """Return _BaseUnweightedCubeCounts subclass instance appropriate to `cube`."""
        dimension_types = cube.dimension_types[-2:]
        valid_counts = cube.unweighted_valid_counts
        counts = valid_counts if valid_counts is not None else cube.unweighted_counts
        diff_nans = True if valid_counts is not None else False
        UnweightedCubeCountsCls = (
            _NumArrayXMrUnweightedCubeCounts
            if dimension_types == (DT.NUM_ARRAY, DT.MR)
            else _NumArrayXCatUnweightedCubeCounts
            if dimension_types[0] == DT.NUM_ARRAY
            else _MrXMrUnweightedCubeCounts
            if dimension_types == (DT.MR, DT.MR)
            else _MrXCatUnweightedCubeCounts
            if dimension_types[0] == DT.MR
            else _CatXMrUnweightedCubeCounts
            if dimension_types[1] == DT.MR
            else _CatXCatUnweightedCubeCounts
        )

        return UnweightedCubeCountsCls(
            dimensions, counts[cls._slice_idx_expr(cube, slice_idx)], diff_nans
        )

    @lazyproperty
    def column_bases(self):
        """2D np.float64 ndarray of column-wise unweighted-N for each matrix cell."""
        return np.broadcast_to(self.columns_base, self.unweighted_counts.shape)

    @lazyproperty
    def columns_base(self):
        """1D or 2D np.float64 ndarray of unweighted column-proportion denominator."""
        raise NotImplementedError(  # pragma: no cover
            "%s must implement `.columns_base`" % type(self).__name__
        )

    @lazyproperty
    def columns_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix column."""
        raise NotImplementedError(  # pragma: no cover
            "%s must implement `.columns_pruning_base`" % type(self).__name__
        )

    @lazyproperty
    def diff_nans(self):
        """Boolean, indicated if subtotal differences (rows and cols) has to be NaN."""
        return self._diff_nans

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of unweighted row-proportion denominator per cell."""
        raise NotImplementedError(  # pragma: no cover
            "%s must implement `.row_bases`" % type(self).__name__
        )

    @lazyproperty
    def rows_base(self):
        """1D or 2D np.float64 ndarray of unweighted row-proportion denominator."""
        raise NotImplementedError(  # pragma: no cover
            "%s must implement `.rows_base`" % type(self).__name__
        )

    @lazyproperty
    def rows_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix row."""
        raise NotImplementedError(  # pragma: no cover
            "%s must implement `.rows_pruning_base`" % type(self).__name__
        )

    @lazyproperty
    def table_base(self):
        """Scalar, 1D, or 2D np.float64 ndarray of unweighted table proportion denom."""
        raise NotImplementedError(  # pragma: no cover
            "%s must implement `.table_base`" % type(self).__name__
        )

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of unweighted table-proportion denominator per cell."""
        raise NotImplementedError(  # pragma: no cover
            "%s must implement `.table_bases`" % type(self).__name__
        )

    @lazyproperty
    def unweighted_counts(self):
        """2D np.float64 ndarray of unweighted-count for each valid matrix cell.

        A valid matrix cell is one whose row and column elements are both non-missing.
        """
        raise NotImplementedError(  # pragma: no cover
            "`%s` must implement `.unweighted_counts`" % type(self).__name__
        )


class _CatXCatUnweightedCubeCounts(_BaseUnweightedCubeCounts):
    """Unweighted-counts cube-measure for a slice with no MR dimensions."""

    @lazyproperty
    def columns_base(self):
        """1D ndarray of np.float64 unweighted-N for each matrix column."""
        return np.sum(self.unweighted_counts, axis=0)

    @lazyproperty
    def columns_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix column.

        Because this matrix has no MR dimension, this is simply the sum of unweighted
        counts for each column.
        """
        return np.sum(self._unweighted_counts, axis=0)

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of unweighted row-proportion denominator per cell."""
        return np.broadcast_to(self.rows_base[:, None], self._unweighted_counts.shape)

    @lazyproperty
    def rows_base(self):
        """1D ndarray of np.float64 unweighted-N for each matrix row."""
        return np.sum(self.unweighted_counts, axis=1)

    @lazyproperty
    def rows_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix row.

        Because this matrix has no MR dimension, this is simply the sum of unweighted
        counts for each row.
        """
        return np.sum(self._unweighted_counts, axis=1)

    @lazyproperty
    def table_base(self):
        """np.float64 count of actual respondents who answered both questions.

        Each dimension of a CAT_X_CAT matrix represents a categorical question. Only
        responses that include answers to both those questions appear as entries in the
        valid elements of those dimensions. The sum total of all valid answers is the
        sample size, aka "N". The term "base" derives from its use as the denominator
        for the table-proportions measure.
        """
        return np.sum(self.unweighted_counts)

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-proportion denominator for each matrix cell."""
        return np.broadcast_to(self.table_base, self._unweighted_counts.shape)

    @lazyproperty
    def unweighted_counts(self):
        """2D np.float64 ndarray of unweighted-count for each valid matrix cell.

        A valid matrix cell is one whose row and column elements are both non-missing.
        """
        return self._unweighted_counts


class _CatXMrUnweightedCubeCounts(_BaseUnweightedCubeCounts):
    """Unweighted-counts cube-measure for a NOT_MR_X_MR slice.

    Note that the rows-dimension need not actually be CAT, as long as it's not MR.
    Its `._unweighted_counts` is a 3D ndarray with axes (rows, cols, selected/not).
    """

    @lazyproperty
    def columns_base(self):
        """2D ndarray of np.float64 unweighted-N for each matrix column."""
        return np.sum(self.unweighted_counts, axis=0)

    @lazyproperty
    def columns_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix column.

        These values include both the selected and unselected counts of the MR columns
        dimension.
        """
        return np.sum(self._unweighted_counts, axis=(0, 2))

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of unweighted row-proportion denominator per cell."""
        # --- in the CAT_X_MR case, rows_base is already the right (2D) value ---
        return self.rows_base

    @lazyproperty
    def rows_base(self):
        """2D np.float64 ndarray of row-wise unweighted-N for this matrix.

        An X_MR matrix has a distinct row-base for each cell. This is because not all
        responses (subvars) are necessarily presented to each respondent. The
        unweighted-count for each X_MR cell is the sum of its selected and unselected
        unweighted counts.
        """
        # --- sel/not axis (2) is summed, rows and columns are preserved ---
        return np.sum(self._unweighted_counts, axis=2)

    @lazyproperty
    def rows_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix row.

        These values include both the selected and unselected counts of the MR columns
        dimension.
        """
        return np.sum(self._unweighted_counts, axis=(1, 2))

    @lazyproperty
    def table_base(self):
        """1D np.float64 unweighted-N for each column of table.

        Because the matrix is X_MR, each column (MR-subvar) has a distinct base.
        """
        # --- unweighted-counts is (nrows, ncols, selected/not) so axis 1 is preserved
        # --- to provide a distinct value for each MR subvar. Both selected and
        # --- not-selected counts contribute to base.
        return np.sum(self._unweighted_counts, axis=(0, 2))

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of unweighted table-proportion denominator per cell."""
        return np.broadcast_to(self.table_base, self.unweighted_counts.shape)

    @lazyproperty
    def unweighted_counts(self):
        """2D np.float64 ndarray of unweighted-count for each valid matrix cell."""
        return self._unweighted_counts[:, :, 0]


class _MrXCatUnweightedCubeCounts(_BaseUnweightedCubeCounts):
    """Unweighted-counts cube-measure for an MR_X_NOT_MR slice.

    Note that the columns-dimension need not actually be CAT, as long as it's not MR.
    Its `._unweighted_counts` is a 3D ndarray with axes (rows, sel/not, cols).
    """

    @lazyproperty
    def columns_base(self):
        """2D np.float64 ndarray of unweighted-N for this matrix.

        An MR_X matrix has a distinct column-base for each cell. This is because not all
        responses (subvars) are necessarily presented to each respondent. The
        unweighted-count for each MR_X cell is the sum of its selected and unselected
        unweighted counts.
        """
        return np.sum(self._unweighted_counts, axis=1)

    @lazyproperty
    def columns_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix column.

        These values include both the selected and unselected counts of the MR rows
        dimension.
        """
        return np.sum(self._unweighted_counts, axis=(0, 1))

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of unweighted row-proportion denominator per cell."""
        return np.broadcast_to(self.rows_base[:, None], self.unweighted_counts.shape)

    @lazyproperty
    def rows_base(self):
        """1D ndarray of np.float64 unweighted-N for each matrix row."""
        # --- only row-selecteds contribute ([:, 0, :]), sum is across columns (axis=1
        # --- after rows sel/not axis is collapsed), rows are retained.
        return np.sum(self._unweighted_counts[:, 0, :], axis=1)

    @lazyproperty
    def rows_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix row.

        These values include both the selected and unselected counts of the MR rows
        dimension.
        """
        return np.sum(self._unweighted_counts, axis=(1, 2))

    @lazyproperty
    def table_base(self):
        """1D np.float64 ndarray (column) of unweighted-N for each row of matrix.

        Since the rows-dimension is MR, each row has a distinct base, since not all of
        the multiple responses were necessarily offered to all respondents. The base for
        each row indicates the number of respondents who were offered that option.
        """
        return np.sum(self._unweighted_counts, axis=(1, 2))

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-proportion denominator for each matrix cell."""
        return np.broadcast_to(self.table_base[:, None], self.unweighted_counts.shape)

    @lazyproperty
    def unweighted_counts(self):
        """2D np.float64 ndarray of unweighted-count for each valid matrix cell."""
        return self._unweighted_counts[:, 0, :]


class _MrXMrUnweightedCubeCounts(_BaseUnweightedCubeCounts):
    """Unweighted-counts cube-measure for an MR_X_MR slice.

    Its `._unweighted_counts` is a 4D ndarray with axes (rows, sel/not, cols, sel/not).
    """

    @lazyproperty
    def columns_base(self):
        """2D np.float64 ndarray of unweighted-N for this matrix.

        An MR_X_MR matrix has a distinct column-base for each cell. This is because not
        all responses (subvars) are necessarily presented to each respondent. The
        unweighted-count for each MR_X cell is the sum of the selected column counts for
        both the selected and unselected row values.
        """
        return np.sum(self._unweighted_counts[:, :, :, 0], axis=1)

    @lazyproperty
    def columns_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix column.

        This includes both selected and unselected counts for the row MR but only
        column-selecteds contribute.
        """
        return np.sum(self._unweighted_counts[:, :, :, 0], axis=(0, 1))

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of unweighted row-proportion denominator per cell."""
        # --- in the MR_X_MR case, rows-base is already the 2D row-unweighted-bases ---
        return self.rows_base

    @lazyproperty
    def rows_base(self):
        """2D np.float64 ndarray of unweighted-N for this matrix.

        An MR_X matrix has a distinct row-base for each cell, the sum of sel-sel and
        sel-not for each cell
        """
        # --- only selecteds in rows contribute ([:, 0, :, :]), selected and not from
        # --- columns both contribute (axis=2 after rows sel/not axis is collapsed),
        # --- both rows and columns are retained, producing a 2D result.
        return np.sum(self._unweighted_counts[:, 0, :, :], axis=2)

    @lazyproperty
    def rows_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix row.

        This includes both selected and unselected counts for the column MR, but only
        selecteds are considered for the rows dimension.
        """
        return np.sum(self._unweighted_counts[:, 0, :, :], axis=(1, 2))

    @lazyproperty
    def table_base(self):
        """2D np.float64 ndarray of distinct unweighted N for each cell of matrix.

        Because the matrix is MR_X_MR, each cell corresponds to a 2x2 sub-table
        (selected/not on each axis), each of which has its own distinct table-base.
        """
        # --- unweighted_counts is 4D of shape (nrows, 2, ncols, 2):
        # --- (MR_SUBVAR (nrows), MR_CAT (sel/not), MR_SUBVAR (ncols), MR_CAT (sel/not))
        # --- Reduce the second and fourth axes with sum() producing 2D (nrows, ncols).
        # --- This sums (selected, selected), (selected, not), (not, selected) and
        # --- (not, not) cells of the subtable for each matrix cell.
        return np.sum(self._unweighted_counts, axis=(1, 3))

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-proportion denominator for each matrix cell."""
        # --- in the MR_X_MR case, table-base is already the 2D table-bases.
        return self.table_base

    @lazyproperty
    def unweighted_counts(self):
        """2D np.float64 ndarray of unweighted-count for each valid matrix cell."""
        # --- indexing is: all-rows, sel-only, all-cols, sel-only ---
        return self._unweighted_counts[:, 0, :, 0]


class _NumArrayXCatUnweightedCubeCounts(_CatXCatUnweightedCubeCounts):
    """Unweighted-counts cube-measure for a slice with no NUM ARRAY row dimension."""

    @lazyproperty
    def column_bases(self):
        """2D np.float64 ndarray of column-wise unweighted-N for each matrix cell."""
        return np.broadcast_to(
            np.sum(self.columns_base, axis=0), self.unweighted_counts.shape
        )

    @lazyproperty
    def columns_base(self):
        """2D ndarray of np.float64 unweighted-N for each matrix column."""
        return self._unweighted_counts


class _NumArrayXMrUnweightedCubeCounts(_CatXMrUnweightedCubeCounts):
    """Unweighted-counts cube-measure for a slice with NUM_ARRAY x MR dimensions."""

    @lazyproperty
    def column_bases(self):
        """2D np.float64 ndarray of column-wise unweighted-N for each matrix cell."""
        return np.broadcast_to(
            np.sum(self.columns_base, axis=0), self.unweighted_counts.shape
        )

    @lazyproperty
    def columns_base(self):
        """2D ndarray of np.float64 unweighted-N for each matrix column."""
        return self._unweighted_counts[:, :, 0]


# === WEIGHTED COUNTS ===


class _BaseWeightedCubeCounts(_BaseCubeMeasure):
    """Base class for weighted-count cube-measure variants."""

    def __init__(self, dimensions, weighted_counts, diff_nans):
        super(_BaseWeightedCubeCounts, self).__init__(dimensions)
        self._weighted_counts = weighted_counts
        self._diff_nans = diff_nans

    @classmethod
    def factory(cls, cube, dimensions, slice_idx):
        """Return _BaseWeightedCounts subclass instance appropriate to `cube`."""
        dimension_types = cube.dimension_types[-2:]
        valid_counts = cube.weighted_valid_counts
        counts = valid_counts if valid_counts is not None else cube.counts
        diff_nans = True if valid_counts is not None else False
        WeightedCubeCountsCls = (
            _MrXMrWeightedCubeCounts
            if dimension_types == (DT.MR, DT.MR)
            else _MrXCatWeightedCubeCounts
            if dimension_types[0] == DT.MR
            else _CatXMrWeightedCubeCounts
            if dimension_types[1] == DT.MR
            else _CatXCatWeightedCubeCounts
        )
        return WeightedCubeCountsCls(
            dimensions, counts[cls._slice_idx_expr(cube, slice_idx)], diff_nans
        )

    @lazyproperty
    def column_bases(self):
        """2D np.float64 ndarray of column-proportion denominator for each cell."""
        return np.broadcast_to(self.columns_margin, self.weighted_counts.shape)

    @lazyproperty
    def columns_margin(self):
        """1D ndarray of np.float64 weighted-N for each matrix column."""
        raise NotImplementedError(  # pragma: no cover
            "%s must implement `.columns_margin`" % type(self).__name__
        )

    @lazyproperty
    def diff_nans(self):
        """Boolean, indicated if subtotal differences (rows and cols) has to be NaN."""
        return self._diff_nans

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of (weighted) row-proportion denominator per cell."""
        raise NotImplementedError(  # pragma: no cover
            "%s must implement `.row_bases`" % type(self).__name__
        )

    @lazyproperty
    def rows_margin(self):
        """1D ndarray of np.float64 weighted-N for each matrix row."""
        raise NotImplementedError(  # pragma: no cover
            "%s must implement `.rows_margin`" % type(self).__name__
        )

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of weighted table-proportion denominator per cell."""
        raise NotImplementedError(  # pragma: no cover
            "%s must implement `.table_bases`" % type(self).__name__
        )

    @lazyproperty
    def table_margin(self):
        """Scalar, 1D, or 2D np.float64 ndarray of weighted-N for matrix.

        There are four cases, one for each of CAT_X_CAT, CAT_X_MR, MR_X_CAT, and
        MR_X_MR. Both CAT_X_MR and MR_X_CAT produce a 1D array, but the former is a
        "row" and the latter is a "column".
        """
        raise NotImplementedError(  # pragma: no cover
            "%s must implement `.table_margin`" % type(self).__name__
        )

    @lazyproperty
    def weighted_counts(self):
        """2D np.float64 ndarray of weighted-count for each valid matrix cell.

        If cube is unweighted, this is the same as unweighted_counts.
        """
        raise NotImplementedError(  # pragma: no cover
            "`%s` must implement `.weighted_counts`" % type(self).__name__
        )

    @lazyproperty
    def zscores(self):
        """2D np.float64 ndarray of zscores for each valid matrix cell."""
        raise NotImplementedError(  # pragma: no cover
            "`%s` must implement `.zscores`" % type(self).__name__
        )

    def _array_type_std_res(self, counts, total, rowsum, colsum):
        """Return 2D np.float64 ndarray of std-res value for each cell of MR matrix.

        This is a utility method used by a matrix with one or more MR dimensions. The
        caller forms the input arrays based on which of its dimensions are MR.
        """
        # --- if the matrix is "defective", in the sense that it doesn't have at least
        # --- two rows and two columns that are "full" of data, don't calculate zscores.
        if not np.all(counts.shape) or np.linalg.matrix_rank(counts) < 2:
            return np.full(counts.shape, np.nan)

        expected_counts = rowsum * colsum / total
        variance = rowsum * colsum * (total - rowsum) * (total - colsum) / total ** 3
        return (counts - expected_counts) / np.sqrt(variance)


class _CatXCatWeightedCubeCounts(_BaseWeightedCubeCounts):
    """Weighted-counts cube-measure for a slice with no MR dimensions."""

    @lazyproperty
    def columns_margin(self):
        """1D ndarray of np.float64 weighted N for each matrix column."""
        return np.sum(self._weighted_counts, axis=0)

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of row-proportion denominator for each matrix cell."""
        return np.broadcast_to(self.rows_margin[:, None], self._weighted_counts.shape)

    @lazyproperty
    def rows_margin(self):
        """1D np.float64 ndarray of weighted-N for each matrix row."""
        return np.sum(self._weighted_counts, axis=1)

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-proportion denominator for each cell."""
        return np.broadcast_to(self.table_margin, self._weighted_counts.shape)

    @lazyproperty
    def table_margin(self):
        """Scalar np.float64 weighted-N for overall table.

        This is the weighted count of respondents who provided a valid response to
        both questions. Because both dimensions are CAT, the table-margin value is the
        same for all cells of the matrix.
        """
        return np.sum(self._weighted_counts)

    @lazyproperty
    def weighted_counts(self):
        """2D np.float64 ndarray of weighted-count for each valid matrix cell.

        If cube is unweighted, this is the same as unweighted_counts.
        """
        return self._weighted_counts

    @lazyproperty
    def zscores(self):
        """2D ndarray of np.float64 std-res value for each cell of matrix.

        A z-score is also known as a *standard score* and is the number of standard
        deviations above (positive) or below (negative) the population mean a cell's
        value is.
        """
        counts = self.weighted_counts

        # --- If the matrix is "defective", in the sense that it doesn't have at least
        # --- two rows and two columns that are "full" of data, don't calculate zscores.
        if not np.all(counts.shape) or np.linalg.matrix_rank(counts) < 2:
            return np.full(counts.shape, np.nan)

        residuals = counts - expected_freq(counts)

        # --- variance of the residuals ---
        rows_margin = self.rows_margin
        columns_margin = self.columns_margin
        table_margin = self.table_margin
        variance_of_residuals = (
            np.outer(rows_margin, columns_margin)
            * np.outer(table_margin - rows_margin, table_margin - columns_margin)
            / table_margin ** 3
        )

        with np.errstate(divide="ignore", invalid="ignore"):
            return residuals / np.sqrt(variance_of_residuals)


class _CatXMrWeightedCubeCounts(_BaseWeightedCubeCounts):
    """Weighted-counts cube-measure for a NOT_MR_X_MR slice.

    Note that the rows-dimension need not actually be CAT, as long as it's not MR.
    Its `._weighted_counts` is a 3D ndarray with axes (rows, cols, selected/not).
    """

    @lazyproperty
    def columns_margin(self):
        """1D ndarray of np.float64 weighted N for each matrix column."""
        # --- only selected counts contribute to the columns margin, which is summed
        # --- across the rows (axis 0).
        return np.sum(self._weighted_counts[:, :, 0], axis=0)

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of row-proportion denominator for each matrix cell."""
        # --- in the X_MR case, row-weighted-bases is the already-2D rows-margin ---
        return self.rows_margin

    @lazyproperty
    def rows_margin(self):
        """2D np.float64 ndarray of weighted-N for each cell of this matrix.

        A matrix with an MR columns dimension has a distinct rows-margin for each cell.
        This is because not all column responses (subvars) are necessarily offered to
        each respondent. The weighted-count for each X_MR cell is the sum of its
        selected and unselected weighted counts.
        """
        # --- selected and not-selected both contribute to margin (axis=2), both rows
        # --- and columns are retained.
        return np.sum(self._weighted_counts, axis=2)

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-proportion denominator for each cell."""
        # --- table-margin is a row-vector (distinct margin for each column), so
        # --- broadcast is vertical
        return np.broadcast_to(self.table_margin, self.weighted_counts.shape)

    @lazyproperty
    def table_margin(self):
        """1D np.float64 ndarray of weighted-N for each column of matrix.

        Because the matrix is X_MR, each column has a distinct table margin.
        """
        # --- weighted-counts is (rows, cols, selected/not) so axis 1 is preserved to
        # --- provide a distinct value for each MR subvar.
        return np.sum(self._weighted_counts, axis=(0, 2))

    @lazyproperty
    def weighted_counts(self):
        """2D np.float64 ndarray of weighted-count for each valid matrix cell.

        If cube is unweighted, this is the same as unweighted_counts.
        """
        return self._weighted_counts[:, :, 0]

    @lazyproperty
    def zscores(self):
        """2D np.float64 ndarray of z-score for each matrix cell."""
        return self._array_type_std_res(
            self.weighted_counts,
            self.table_margin,
            np.sum(self._weighted_counts, axis=2),
            np.sum(self.weighted_counts, axis=0),
        )


class _MrXCatWeightedCubeCounts(_BaseWeightedCubeCounts):
    """Weighted-counts cube-measure for an MR_X_NOT_MR slice.

    Its `._weighted_counts` is a 3D ndarray with axes (rows, sel/not, cols).
    """

    @lazyproperty
    def columns_margin(self):
        """2D np.float64 ndarray of weighted-N for each cell of this matrix.

        An MR_X matrix has a distinct column-margin for each cell. This is because not
        all responses (subvars) are necessarily presented to each respondent. The
        weighted-count for each MR_X cell is the sum of its selected and unselected
        weighted counts.
        """
        # --- sum sel and not-sel (axis 1), rows and columns are retained ---
        return np.sum(self._weighted_counts, axis=1)

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of row-proportion denominator for each matrix cell."""
        return np.broadcast_to(self.rows_margin[:, None], self.weighted_counts.shape)

    @lazyproperty
    def rows_margin(self):
        """1D np.float64 ndarray of weighted-N for each matrix row."""
        return np.sum(self.weighted_counts, axis=1)

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-proportion denominator for each cell."""
        return np.broadcast_to(self.table_margin[:, None], self.weighted_counts.shape)

    @lazyproperty
    def table_margin(self):
        """1D np.float64 ndarray of weighted-N for each row of matrix.

        Since the rows-dimension is MR, each row has a distinct table margin, since not
        all of the multiple responses were necessarily offered to all respondents. The
        table-margin for each row indicates the weighted number of respondents who were
        offered that option.
        """
        return np.sum(self._weighted_counts, axis=(1, 2))

    @lazyproperty
    def weighted_counts(self):
        """2D np.float64 ndarray of weighted-count for each valid matrix cell.

        If cube is unweighted, this is the same as unweighted_counts.
        """
        return self._weighted_counts[:, 0, :]

    @lazyproperty
    def zscores(self):
        """2D np.float64 ndarray of z-score for each matrix cell."""
        return self._array_type_std_res(
            self.weighted_counts,
            self.table_margin[:, np.newaxis],
            np.sum(self.weighted_counts, axis=1)[:, None],
            np.sum(self._weighted_counts, axis=1),
        )


class _MrXMrWeightedCubeCounts(_BaseWeightedCubeCounts):
    """Weighted-counts cube-measure for an MR_X_MR slice.

    Its `._weighted_counts` is a 4D ndarray with axes (rows, sel/not, cols, sel/not).
    """

    @lazyproperty
    def columns_margin(self):
        """2D np.float64 ndarray of weighted-N for each cell of matrix.

        An MR_X matrix has a distinct columns-margin for each cell. This is because not
        all responses (subvars) are necessarily presented to each respondent. Each
        MR_X_MR cell has four counts: sel-sel, sel-not, not-sel, and not-not. Only
        sel-sel and not-sel contribute to the columns-margin.
        """
        # --- only column-selected counts contribute ([:, :, :, 0]), row-selected and
        # --- not-selected are summed (axis=1), rows and columns are retained.
        return np.sum(self._weighted_counts[:, :, :, 0], axis=1)

    @lazyproperty
    def row_bases(self):
        """2D np.float64 ndarray of row-proportion denominator for each matrix cell."""
        # --- in the X_MR case, row-margins is the already-2D rows-margin ---
        return self.rows_margin

    @lazyproperty
    def rows_margin(self):
        """2D np.float64 ndarray of weighted-N for each cell of matrix.

        An X_MR matrix has a distinct rows-margin for each cell. Each MR_X_MR cell has
        four counts: sel-sel, sel-not, not-sel, and not-not. Only sel-sel and sel-not
        contribute to the rows-margin.
        """
        # --- only selecteds in rows contribute ([:, 0, :, :]), selected and not from
        # --- columns both contribute (axis=2 after rows sel/not axis is collapsed),
        # --- both rows and columns are retained, producing a 2D result.
        return np.sum(self._weighted_counts[:, 0, :, :], axis=2)

    @lazyproperty
    def table_bases(self):
        """2D np.float64 ndarray of table-proportion denominator for each cell."""
        # --- in MR_X_MR case, table-bases is the already-2D table-margin ---
        return self.table_margin

    @lazyproperty
    def table_margin(self):
        """2D np.float64 ndarray of weighted-N for each cell of matrix.

        Because the matrix is MR_X_MR, each cell corresponds to a 2x2 sub-table
        (selected/not on each axis), each of which has its own distinct table-margin.
        """
        # --- Reduce second and fourth axes (the two MR_CAT dimensions) with sum()
        # --- producing 2D (nrows, ncols). This sums the (selected, selected),
        # --- (selected, not), (not, selected) and (not, not) cells of the subtable for
        # --- each matrix cell. Rows and columns are retained.
        return np.sum(self._weighted_counts, axis=(1, 3))

    @lazyproperty
    def weighted_counts(self):
        """2D np.float64 ndarray of weighted-count for each valid matrix cell.

        If cube is unweighted, this is the same as unweighted_counts. Only *selected*
        counts contribute to these values.
        """
        return self._weighted_counts[:, 0, :, 0]

    @lazyproperty
    def zscores(self):
        """2D ndarray of np.float64 std-res value for each cell of matrix.

        A z-score is also known as a *standard score* and is the number of standard
        deviations above (positive) or below (negative) the population mean each cell's
        value is.
        """
        return self._array_type_std_res(
            self.weighted_counts,
            self.table_margin,
            np.sum(self._weighted_counts, axis=3)[:, 0, :],
            np.sum(self._weighted_counts, axis=1)[:, :, 0],
        )


# === LEGACY MATRIX OBJECTS ===


class BaseCubeResultMatrix(object):
    """Base class for all cube-result matrix (2D second-order analyzer) objects."""

    def __init__(
        self, dimensions, weighted_counts, unweighted_counts, counts_with_missings=None
    ):
        self._dimensions = dimensions
        self._weighted_counts = weighted_counts
        self._unweighted_counts = unweighted_counts
        self._counts_with_missings = counts_with_missings

    @classmethod
    def factory(cls, cube, dimensions, slice_idx):
        """Return a base-matrix object of appropriate type for `cube`."""
        dimension_types = cube.dimension_types[-2:]
        MatrixCls = (
            _MrXMrMatrix
            if dimension_types == (DT.MR, DT.MR)
            else _MrXCatMatrix
            if dimension_types[0] == DT.MR
            else _CatXMrMatrix
            if dimension_types[1] == DT.MR
            else _CatXCatMatrix
        )
        return MatrixCls(dimensions, *cls._sliced_counts(cube, slice_idx))

    @lazyproperty
    def column_index(self):
        """2D np.float64/np.nan ndarray of column-index value for each matrix cell.

        Column-index answers the question "are respondents in this row-category more or
        less likely than the overall table population to choose the answer represented
        by this column?". For example, if the row is "Hispanic" and the column is
        home-ownership, a value of 100 indicates hispanics are no less and no more
        likely to own their home than the overall population. If that value was 150, it
        would indicate hispanics are 50% more likely to own their home than the general
        population (or the population surveyed anyway).
        """
        return self.column_proportions / self._baseline * 100

    @lazyproperty
    def column_proportions(self):
        """2D ndarray of np.float64 between 0.0 and 1.0.

        The value represents the ratio of each cell count to the total count (margin)
        for its column.
        """
        return self.weighted_counts / self.columns_margin

    @lazyproperty
    def columns_dimension(self):
        """The `Dimension` object representing column elements of this matrix."""
        return self._dimensions[1]

    @lazyproperty
    def columns_margin(self):
        """1D/2D np.float64 ndarray of weighted-N for each column of matrix."""
        raise NotImplementedError(
            "`%s` must implement `.columns_margin`" % type(self).__name__
        )  # pragma: no cover

    @lazyproperty
    def columns_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix column."""
        raise NotImplementedError(
            "`%s` must implement `.columns_pruning_base`" % type(self).__name__
        )  # pragma: no cover

    @lazyproperty
    def rows_base(self):
        """1D/2D np.float64 ndarray of unweighted-N for each matrix row/cell."""
        raise NotImplementedError(
            "`%s` must implement `.rows_base`" % type(self).__name__
        )  # pragma: no cover

    @lazyproperty
    def rows_dimension(self):
        """The `Dimension` object representing row elements of this matrix."""
        return self._dimensions[0]

    @lazyproperty
    def rows_margin(self):
        """1D/2D np.float64 ndarray of weighted-N for each matrix row/cell."""
        raise NotImplementedError(
            "`%s` must implement `.rows_margin`" % type(self).__name__
        )  # pragma: no cover

    @lazyproperty
    def rows_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix row."""
        raise NotImplementedError(
            "`%s` must implement `.rows_pruning_base`" % type(self).__name__
        )  # pragma: no cover

    @lazyproperty
    def table_base(self):
        """Scalar, 1D, or 2D ndarray of np.float64 unweighted-N for this slice.

        This value has four distinct forms, depending on the subclass.
        """
        raise NotImplementedError(
            "`%s` must implement `.table_base`" % type(self).__name__
        )  # pragma: no cover

    @lazyproperty
    def table_margin(self):
        """np.float64 scalar or a 1D or 2D np.float64 ndarray table margin.

        The table margin is the overall sample size of the matrix. This is the weighted
        count of respondents who were asked both questions and provided a valid response
        for both (including not-selecting an MR option/subvar).

        A matrix with a multiple-response (MR) dimension produces a 1D ndarray value.
        When both dimensions are MR, the return value is a 2D ndarray and there is
        a distinct table-base value for each "cell" of the matrix. A CAT_X_CAT matrix
        produces a scalar value for this property.
        """
        raise NotImplementedError(
            "`%s` must implement `.table_margin" % self.__class__.__name__
        )  # pragma: no cover

    @lazyproperty
    def table_stderrs(self):
        """2D np.float64 ndarray of table-percent std-error for each matrix cell."""
        raise NotImplementedError(
            "`%s` must implement `.table_stderrs" % self.__class__.__name__
        )  # pragma: no cover

    @lazyproperty
    def unweighted_counts(self):
        """2D np.float64 ndarray of unweighted-count for each valid matrix cell."""
        raise NotImplementedError(
            "`%s` must implement `.unweighted_counts" % type(self).__name__
        )  # pragma: no cover

    @lazyproperty
    def weighted_counts(self):
        """2D np.float64 ndarray of weighted-count for each valid matrix cell.

        If cube is unweighted, this is the same as unweighted_counts.
        """
        raise NotImplementedError(
            "`%s` must implement `.weighted_counts`" % type(self).__name__
        )  # pragma: no cover

    @staticmethod
    def _cube_slice_expression(cube, slice_idx):
        """return `np.s_` object with correct slicing for the cube type."""
        if cube.ndim <= 2:
            return np.s_[:]

        # --- If 0th dimension of a >2D cube is MR, we only need the "Selected"
        # --- counts, because it's "just" used to tabulate.
        if cube.dimension_types[0] == DT.MR:
            return np.s_[slice_idx, 0]

        # --- If we have a cube with more than 2 dimensions we need to extract the
        # --- appropriate slice (element of the 0th dimension).
        return np.s_[slice_idx]

    @classmethod
    def _sliced_counts(cls, cube, slice_idx):
        """Return tuple of cube counts, prepared for regular matrix construction.

        Depending on the type of the cube, we need to extract the proper counts for the
        counstruction of a particular slice (matrix). In case of cubes that have more
        then 2 dimensions, we only need a particular slice (a particular selected
        element of the 0th dimension).

        If, in addition to being >2D cube, the 0th dimension is multiple response, we
        need to extract only the selected counts, since we're "just" dealing with the
        tabulation.
        """
        slice_expr = cls._cube_slice_expression(cube, slice_idx)
        return (
            cube.counts[slice_expr],
            cube.unweighted_counts[slice_expr],
            cube.counts_with_missings[slice_expr],
        )

    @lazyproperty
    def _valid_row_idxs(self):
        """ndarray-style index for only valid (non-missing) rows.

        Suitable for indexing a raw measure array to include only valid rows.
        """
        return np.ix_(self._dimensions[-2].valid_elements.element_idxs)


class _CatXCatMatrix(BaseCubeResultMatrix):
    """Matrix for CAT_X_CAT cubes and base class for most other matrix classes.

    Despite the name, this matrix is used for CA_SUBVAR and CA_CAT dimension too, since
    these behave the same from a base-matrix perspective.

    `counts_with_missings` is the raw weighted counts array, needed to compute the
    column-index.
    """

    @lazyproperty
    def column_index(self):
        """2D np.float64/np.nan ndarray of column-index value for each matrix cell.

        Column-index answers the question "are respondents in this row-category more or
        less likely than the overall table population to choose the answer represented
        by this column?". For example, if the row is "Hispanic" and the column is
        home-ownership, a value of 100 indicates hispanics are no less and no more
        likely to own their home than the overall population. If that value was 150, it
        would indicate hispanics are 50% more likely to own their home than the general
        population (or the population surveyed anyway).
        """
        return self.column_proportions / self._baseline * 100

    @lazyproperty
    def columns_margin(self):
        """1D ndarray of np.float64 weighted N for each matrix column."""
        return np.sum(self.weighted_counts, axis=0)

    @lazyproperty
    def columns_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix column.

        Because this matrix has no MR dimension, this is simply the sum of unweighted
        counts for each column.
        """
        return np.sum(self._unweighted_counts, axis=0)

    @lazyproperty
    def rows_base(self):
        """1D ndarray of np.float64 unweighted-N for each matrix row."""
        return np.sum(self.unweighted_counts, axis=1)

    @lazyproperty
    def rows_margin(self):
        """1D np.float64 ndarray of weighted-N for each matrix row."""
        return np.sum(self._weighted_counts, axis=1)

    @lazyproperty
    def rows_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix row.

        Because this matrix has no MR dimension, this is simply the sum of unweighted
        counts for each row.
        """
        return np.sum(self._unweighted_counts, axis=1)

    @lazyproperty
    def table_base(self):
        """np.float64 count of actual respondents who answered both questions.

        Each dimension of a CAT_X_CAT matrix represents a categorical question. Only
        responses that include answers to both those questions appear as entries in the
        valid elements of those dimensions. The sum total of all valid answers is the
        sample size, aka "N" or "base".
        """
        return np.sum(self.unweighted_counts)

    @lazyproperty
    def table_margin(self):
        """Scalar np.float64 weighted-N for overall table.

        This is the weighted count of respondents who provided a valid response to
        both questions. Because both dimensions are CAT, the table-margin value is the
        same for all cells of the matrix.
        """
        return np.sum(self._weighted_counts)

    @lazyproperty
    def table_stderrs(self):
        """2D np.float64 ndarray of table-percent std-error for each matrix cell.

        Standard error is sqrt(variance/N).
        """
        return np.sqrt(self._table_proportion_variances / self.table_margin)

    @lazyproperty
    def unweighted_counts(self):
        """2D np.float64 ndarray of unweighted-count for each valid matrix cell.

        A valid matrix cell is one whose row and column elements are both non-missing.
        """
        return self._unweighted_counts

    @lazyproperty
    def weighted_counts(self):
        """2D np.float64 ndarray of weighted-count for each valid matrix cell.

        If cube is unweighted, this is the same as unweighted_counts.
        """
        return self._weighted_counts

    @lazyproperty
    def _baseline(self):
        """2D np.float64 ndarray of baseline value for each row in matrix.

        The shape of the return value is (nrows, 1). The baseline value for a row is the
        proportion of all values that appear in that row. A baseline for a 4 x 3 matrix
        looks like this:

            [[0.2006734 ]
             [0.72592593]
             [0.05521886]
             [0.01818182]]

        Note that the baseline values sum to 1.0. This is because each represents the
        portion of all responses that fall in that row. This baseline value is the
        denominator of the `._column_index` computation.

        Baseline is a straightforward function of the *unconditional row margin*.
        Unconditional here means that both valid and invalid responses (to the
        columns-var question) are included. This ensures that the baseline is not
        distorted by a large number of missing responses to the columns-question.
        """
        # --- uncond_row_margin is a 1D ndarray of the weighted total observation count
        # --- involving each valid row. Counts consider both valid and invalid columns,
        # --- but are only produced for valid rows.
        uncond_row_margin = np.sum(self._counts_with_missings, axis=1)[
            self._valid_row_idxs
        ]
        return uncond_row_margin[:, None] / np.sum(uncond_row_margin)

    @lazyproperty
    def _table_proportion_variances(self):
        """2D ndarray of np.float64 cell proportion variance for each cell of matrix."""
        p = self._weighted_counts / self.table_margin
        return p * (1 - p)


class _CatXMrMatrix(_CatXCatMatrix):
    """Represents a CAT x MR slice.

    Its `._counts` is a 3D ndarray of np.float64 with axes (rows, cols, selected/not),
    like:

        [[[1002.52343241 1247.791605  ]
          [ 765.95079804 1484.36423937]
          [ 656.43937497 1593.87566244]]

         [[1520.23482091 2573.22762247]
          [1291.0925792  2802.36986418]
          [1595.44412365 2498.01831973]]

         [[ 908.65667501 2254.62623471]
          [ 841.76439186 2321.51851785]
          [1603.79596755 1559.48694217]]

         [[ 746.89008236 1753.26322241]
          [ 721.38248086 1778.7708239 ]
          [1255.87038944 1244.28291533]]

         [[   9.83166357   25.9551254 ]
          [   8.23140253   27.55538645]
          [  22.214956     13.57183298]]]
    """

    @lazyproperty
    def columns_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix column.

        These values include both the selected and unselected counts of the MR columns
        dimension.
        """
        return np.sum(self._unweighted_counts, axis=(0, 2))

    @lazyproperty
    def rows_base(self):
        """2D np.float64 ndarray of row-wise unweighted-N for this matrix.

        An X_MR matrix has a distinct row-base for each cell. This is because not all
        responses (subvars) are necessarily presented to each respondent. The
        unweighted-count for each X_MR cell is the sum of its selected and unselected
        unweighted counts.
        """
        return np.sum(self._unweighted_counts, axis=2)

    @lazyproperty
    def rows_margin(self):
        """2D np.float64 ndarray of weighted-N for each cell of this matrix.

        A matrix with an MR columns dimension has a distinct rows-margin for each cell.
        This is because not all column responses (subvars) are necessarily offered to
        each respondent. The weighted-count for each X_MR cell is the sum of its
        selected and unselected weighted counts.
        """
        return np.sum(self._weighted_counts, axis=2)

    @lazyproperty
    def rows_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix row.

        These values include both the selected and unselected counts of the MR columns
        dimension.
        """
        return np.sum(self._unweighted_counts, axis=(1, 2))

    @lazyproperty
    def table_base(self):
        """1D np.float64 unweighted N for each column of matrix.

        Because the matrix is X_MR, each column has a distinct table base.
        """
        # --- unweighted-counts is (nrows, ncols, selected/not) so axis 1 is preserved
        # --- to provide a distinct value for each MR subvar.
        return np.sum(self._unweighted_counts, axis=(0, 2))

    @lazyproperty
    def table_margin(self):
        """1D np.float64 ndarray of weighted-N for each column of matrix.

        Because the matrix is X_MR, each column has a distinct table margin.
        """
        # --- weighted-counts is (rows, cols, selected/not) so axis 1 is preserved to
        # --- provide a distinct value for each MR subvar.
        return np.sum(self._weighted_counts, axis=(0, 2))

    @lazyproperty
    def unweighted_counts(self):
        """2D np.float64 ndarray of unweighted-count for each valid matrix cell.

        A valid matrix cell is one whose row and column elements are both non-missing.
        """
        return self._unweighted_counts[:, :, 0]

    @lazyproperty
    def weighted_counts(self):
        """2D np.float64 ndarray of weighted-count for each valid matrix cell.

        A valid matrix cell is one whose row and column elements are both non-missing.
        If the cube-result has no weight, these values are the same as the
        unweighted-counts.
        """
        return self._weighted_counts[:, :, 0]

    @lazyproperty
    def _baseline(self):
        """2D np.float64 (or NaN) ndarray of baseline value for each matrix cell.

        Its shape is (nrows, ncols) which corresponds to CAT_X_MR_SUBVAR.

        The baseline value is compared with the column-proportion value for each cell to
        form the column-index value. The baseline is a function of the unconditional row
        margin, which is the sum of counts across both valid and missing columns.

        For CAT_X_MR, `uncond_row_margin` sums across the MR_CAT (selected, not,
        missing) dimension to include missing values (an MR_SUBVAR element is never
        "missing": true).
        """
        # --- counts_with_missings.shape is (nall_rows, ncols, selected/not/missing).
        # --- axes[1] corresponds to the MR_SUBVAR dimension, in which there are never
        # --- "missing" subvars (so nall_cols always equals ncols for that dimension
        # --- type). uncond_row_margin selects only valid rows, retains all columns and
        # --- reduces the selected/not/missing axis by summing those three counts. Its
        # --- shape is (nrows, ncols).
        uncond_row_margin = np.sum(self._counts_with_missings, axis=2)[
            self._valid_row_idxs
        ]
        # --- uncond_table_margin sums across rows, producing 1D array of size ncols,
        # --- (although all its values are always the same).
        uncond_table_margin = np.sum(uncond_row_margin, axis=0)
        # --- division produces a 2D matrix of shape (nrows, ncols) ---
        return uncond_row_margin / uncond_table_margin

    @lazyproperty
    def _table_proportion_variances(self):
        """2D ndarray of np.float64 table proportion variance for each matrix cell."""
        p = self._weighted_counts[:, :, 0] / self.table_margin
        return p * (1 - p)


class _MrXCatMatrix(BaseCubeResultMatrix):
    """Represents an MR_X_CAT slice.

    Its `._counts` is a 3D ndarray of np.float64 with axes (rows, sel/not, cols), like:

        [[[ 39  44  24  35]
          [389 447 266 394]]

         [[ 34  36  29  24]
          [394 455 261 405]]

         [[357 415 241 371]
          [ 71  76  49  58]]

         [[  0   0   0   0]
          [428 491 290 429]]]
    """

    @lazyproperty
    def columns_margin(self):
        """2D np.float64 ndarray of weighted-N for each cell of this matrix.

        An MR_X matrix has a distinct column-margin for each cell. This is because not
        all responses (subvars) are necessarily presented to each respondent. The
        weighted-count for each MR_X cell is the sum of its selected and unselected
        weighted counts.
        """
        return np.sum(self._weighted_counts, axis=1)

    @lazyproperty
    def columns_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix column.

        These values include both the selected and unselected counts of the MR rows
        dimension.
        """
        return np.sum(self._unweighted_counts, axis=(0, 1))

    @lazyproperty
    def rows_base(self):
        """1D ndarray of np.float64 unweighted-N for each matrix row.

        Only selected values contribute to the rows-base.
        """
        return np.sum(self._unweighted_counts[:, 0, :], axis=1)

    @lazyproperty
    def rows_margin(self):
        """1D np.float64 ndarray of weighted-N for each matrix row.

        Only selected values contribute to the rows-margin.
        """
        return np.sum(self._weighted_counts[:, 0, :], axis=1)

    @lazyproperty
    def rows_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix row.

        These values include both the selected and unselected counts of the MR rows
        dimension.
        """
        return np.sum(self._unweighted_counts, axis=(1, 2))

    @lazyproperty
    def table_base(self):
        """1D np.float64 ndarray of unweighted N for each row of matrix.

        Since the rows-dimension is MR, each row has a distinct base, since not all of
        the multiple responses were necessarily offered to all respondents. The base for
        each row indicates the number of respondents who were offered that option.
        """
        return np.sum(self._unweighted_counts, axis=(1, 2))

    @lazyproperty
    def table_margin(self):
        """1D np.float64 ndarray of weighted-N for each row of matrix.

        Since the rows-dimension is MR, each row has a distinct table margin, since not
        all of the multiple responses were necessarily offered to all respondents. The
        table-margin for each row indicates the weighted number of respondents who were
        offered that option.
        """
        return np.sum(self._weighted_counts, axis=(1, 2))

    @lazyproperty
    def table_stderrs(self):
        """2D np.float64 ndarray of table-percent std-error for each matrix cell.

        Standard error is sqrt(variance/N).
        """
        return np.sqrt(self._table_proportion_variances / self.table_margin[:, None])

    @lazyproperty
    def unweighted_counts(self):
        """2D np.float64 ndarray of unweighted-count for each valid matrix cell.

        A valid matrix cell is one whose row and column elements are both non-missing.
        """
        return self._unweighted_counts[:, 0, :]

    @lazyproperty
    def weighted_counts(self):
        """2D np.float64 ndarray of weighted-count for each valid matrix cell.

        A valid matrix cell is one whose row and column elements are both non-missing.
        When the cube-result has no weight, in which case these values are the same as
        the unweighted-counts.
        """
        return self._weighted_counts[:, 0, :]

    @lazyproperty
    def _baseline(self):
        """2D np.float64 ndarray of baseline value for each row in matrix.

        `._baseline` is the denominator of the column-index and represents the
        proportion of the overall row-count present in each row. A cell with
        a column-proportion exactly equal to this basline will have a column-index of
        100.

        The shape of the return value is (nrows, 1). A baseline for a 4 x 3 matrix looks
        something like this:

            [[0.17935204]
             [0.33454989]
             [0.50762388]
             [0.80331259]
             [0.7996507 ]]

        Baseline is a function of the *unconditional row margin*. Unconditional here
        means that both valid and invalid responses (to the columns-var question) are
        included. This ensures that the baseline is not distorted by a large number of
        missing responses to the columns-question.
        """
        # --- unconditional row-margin is a 1D ndarray of size nrows computed by:
        # --- 1. summing across all columns: np.sum(self._counts_with_missings, axis=2)
        # --- 2. taking only selected counts: [:, 0]
        # --- 3. taking only valid rows: [self._valid_row_idxs]
        uncond_row_margin = np.sum(self._counts_with_missings, axis=2)[:, 0][
            self._valid_row_idxs
        ]
        # --- The "total" (uncond_row_table_margin) is a 1D ndarray of size nrows. Each
        # --- sum includes only valid rows (MR_SUBVAR, axis 0), selected and unselected
        # --- but not missing counts ([0:2]) of the MR_CAT axis (axis 1), and all column
        # --- counts, both valid and missing (axis 2). The rows axis (0) is preserved
        # --- because each MR subvar has a distinct table margin.
        uncond_row_table_margin = np.sum(
            self._counts_with_missings[self._valid_row_idxs][:, 0:2], axis=(1, 2)
        )
        # --- inflate shape to (nrows, 1) for later calculation convenience ---
        return (uncond_row_margin / uncond_row_table_margin)[:, None]

    @lazyproperty
    def _table_proportion_variances(self):
        """2D ndarray of np.float64 table proportion variance for each matrix cell."""
        p = self._weighted_counts[:, 0, :] / self.table_margin[:, None]
        return p * (1 - p)


class _MrXMrMatrix(_CatXCatMatrix):
    """Represents an MR x MR slice.

    Its `._counts` is a 4D ndarray with axes (rows, sel/not, cols, sel/not), like:

        [[[[2990.03485848 4417.96127006]
           [2713.94318797 4694.05294056]
           [2847.96860219 4560.02752634]]

          [[1198.10181578 3436.90253993]
           [ 914.47846452 3720.52589119]
           [2285.79620941 2349.2081463 ]]]


         [[[2626.08325048 5180.55485426]
           [2396.04310657 5410.59499817]
           [3503.08635211 4303.55175262]]

          [[1562.05342378 2674.30895573]
           [1232.37854592 3003.98383359]
           [1630.67845949 2605.68392002]]]


         [[[3370.04923406 5278.54391705]
           [3033.71862569 5614.87452542]
           [3312.56140096 5336.03175016]]

          [[ 818.0874402  2576.31989293]
           [ 594.7030268  2799.70430633]
           [1821.20341065 1573.20392249]]]


         [[[1822.67560537 2883.99243344]
           [1616.70492531 3089.96311351]
           [1735.59793395 2971.07010487]]

          [[2365.46106889 4970.87137654]
           [2011.71672718 5324.61571825]
           [3398.16687766 3938.16556777]]]]

    """

    @lazyproperty
    def columns_margin(self):
        """2D np.float64 ndarray of weighted-N for each cell of matrix.

        An MR_X matrix has a distinct columns-margin for each cell. This is because not
        all responses (subvars) are necessarily presented to each respondent. Each
        MR_X_MR cell has four counts: sel-sel, sel-not, not-sel, and not-not. Only
        sel-sel and not-sel contribute to the columns-margin.
        """
        return np.sum(self._weighted_counts[:, :, :, 0], axis=1)

    @lazyproperty
    def columns_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix column.

        Because both dimensions of this matrix are MR, this includes both selected and
        unselected counts, but only for the row MR; only selecteds are considered for
        the columns dimension.
        """
        return np.sum(self._unweighted_counts[:, :, :, 0], axis=(0, 1))

    @lazyproperty
    def rows_base(self):
        """2D np.float64 ndarray of unweighted-N for this matrix.

        An MR_X matrix has a distinct row-base for each cell, the sum of sel-sel and
        sel-not for each cell
        """
        return np.sum(self._unweighted_counts[:, 0, :, :], axis=2)

    @lazyproperty
    def rows_margin(self):
        """2D np.float64 ndarray of weighted-N for each cell of matrix.

        An X_MR matrix has a distinct rows-margin for each cell. Each MR_X_MR cell has
        four counts: sel-sel, sel-not, not-sel, and not-not. Only sel-sel and sel-not
        contribute to the rows-margin.
        """
        # --- sum of (sel-sel, sel-not) ---
        return np.sum(self._weighted_counts[:, 0, :, :], axis=2)

    @lazyproperty
    def rows_pruning_base(self):
        """1D np.float64 ndarray of unweighted-N for each matrix row.

        Because both dimensions of this matrix are MR, this includes both selected and
        unselected counts, but only for the column MR; only selecteds are considered for
        the rows dimension.
        """
        return np.sum(self._unweighted_counts[:, 0, :, :], axis=(1, 2))

    @lazyproperty
    def table_base(self):
        """2D np.float64 ndarray of distinct unweighted N for each cell of matrix.

        Because the matrix is MR_X_MR, each cell corresponds to a 2x2 sub-table
        (selected/not on each axis), each of which has its own distinct table-base.
        """
        # --- unweighted_counts is 4D of shape (nrows, 2, ncols, 2):
        # --- (MR_SUBVAR (nrows), MR_CAT (sel/not), MR_SUBVAR (ncols), MR_CAT (sel/not))
        # --- Reduce the second and fourth axes with sum() producing 2D (nrows, ncols).
        # --- This sums (selected, selected), (selected, not), (not, selected) and
        # --- (not, not) cells of the subtable for each matrix cell.
        return np.sum(self._unweighted_counts, axis=(1, 3))

    @lazyproperty
    def table_margin(self):
        """2D np.float64 ndarray of weighted N for each cell of matrix.

        Because the matrix is MR_X_MR, each cell corresponds to a 2x2 sub-table
        (selected/not on each axis), each of which has its own distinct table-margin.
        """
        # --- Reduce second and fourth axes (the two MR_CAT dimensions) with sum()
        # --- producing 2D (nrows, ncols). This sums the (selected, selected),
        # --- (selected, not), (not, selected) and (not, not) cells of the subtable for
        # --- each matrix cell.
        return np.sum(self._weighted_counts, axis=(1, 3))

    @lazyproperty
    def unweighted_counts(self):
        """2D np.float64 ndarray of unweighted-count for each valid matrix cell.

        A valid matrix cell is one whose row and column elements are both non-missing.
        """
        return self._unweighted_counts[:, 0, :, 0]

    @lazyproperty
    def weighted_counts(self):
        """2D np.float64 ndarray of weighted-count for each valid matrix cell.

        When the cube-result has no weight, these values are the same as the
        unweighted-counts. Only *selected* counts contribute to these values.
        """
        return self._weighted_counts[:, 0, :, 0]

    @lazyproperty
    def _baseline(self):
        """2D np.float64 ndarray of baseline value for each matrix cell.

        The shape is (nrows, ncols) and all values in a given row are the same. So
        really there are only nrows distinct baseline values, but the returned shape
        makes calculating column-index in a general way more convenient.
        """
        # --- `counts_with_missings` for MR_X_MR is 4D of size (nrows, 3, ncols, 3)
        # --- (MR_SUBVAR, MR_CAT, MR_SUBVAR, MR_CAT). Unconditional row margin:
        # --- * Takes all rows and all cols (axes 0 & 2), because MR_SUBVAR dimension
        # ---   can contain only valid elements (no such thing as "missing": true
        # ---   subvar).
        # --- * Sums selected + unselected + missing categories in second MR_CAT
        # ---   dimension (columns MR, axes[3]). Including missings here fulfills
        # ---   "unconditional" characteristic of margin.
        # --- * Takes only those totals associated with selected categories of first
        # ---    MR_CAT dimension (rows MR). ("counts" for MR are "selected" counts).
        # --- Produces a 2D (nrows, ncols) array.
        uncond_row_margin = np.sum(self._counts_with_missings[:, 0:2], axis=3)[:, 0]
        # --- Unconditional table margin is also 2D (nrows, ncols) but the values for
        # --- all columns in a row have the same value; basically each row has
        # --- a distinct table margin.
        uncond_table_margin = np.sum(self._counts_with_missings[:, 0:2], axis=(1, 3))
        # --- baseline is produced by dividing uncond_row_margin by uncond_table_margin.
        return uncond_row_margin / uncond_table_margin

    @lazyproperty
    def _table_proportion_variances(self):
        """2D ndarray of np.float64 table proportion variance for each matrix cell."""
        p = self._weighted_counts[:, 0, :, 0] / self.table_margin
        return p * (1 - p)



# encoding: utf-8

"""Partial-integration test suite for `cr.cube.collator` module."""

import pytest

from cr.cube.collator import (
    ExplicitOrderCollator,
    PayloadOrderCollator,
    SortByValueCollator,
)
from cr.cube.dimension import Dimension, _OrderSpec, _Subtotal
from cr.cube.enums import DIMENSION_TYPE as DT

from ..unitutil import instance_mock


class DescribeExplicitOrderCollator(object):
    """Partial-integration test suite for `ExplicitOrderCollator` object."""

    @pytest.mark.parametrize(
        "element_ids, element_subvar_ids, order, anchors, expected_value",
        (
            (
                (1, 2, 3),
                [None, None, None],
                [2, 3, 1],
                ("bottom", 3, 3, "top"),
                (-1, 1, 2, -3, -2, 0, -4),
            ),
            ((9, 3, 7), [None, None, None], [7], (), (2, 0, 1)),
        ),
    )
    def it_knows_the_display_order_for_a_dimension(
        self, request, element_ids, element_subvar_ids, order, anchors, expected_value
    ):
        subtotals_ = [instance_mock(request, _Subtotal, anchor=a) for a in anchors]
        order_spec_ = instance_mock(request, _OrderSpec, element_ids=order)
        dimension_ = instance_mock(
            request,
            Dimension,
            element_ids=element_ids,
            element_subvar_ids=element_subvar_ids,
            subtotals=subtotals_,
            order_spec=order_spec_,
        )

        assert ExplicitOrderCollator.display_order(dimension_, ()) == expected_value


class DescribePayloadOrderCollator(object):
    """Partial-integration test suite for `PayloadOrderCollator` object."""

    @pytest.mark.parametrize(
        "element_ids, anchors, expected_value",
        (
            ((9, 3, 7), ("bottom", 3, 3, "top"), (-1, 0, 1, -3, -2, 2, -4)),
            ((9, 3, 7), (), (0, 1, 2)),
            ((), ("bottom", 3, 3, "top"), (-1, -4, -3, -2)),
        ),
    )
    def it_knows_the_display_order_for_a_dimension(
        self, request, element_ids, anchors, expected_value
    ):
        subtotals_ = [instance_mock(request, _Subtotal, anchor=a) for a in anchors]
        dimension_ = instance_mock(
            request, Dimension, element_ids=element_ids, subtotals=subtotals_
        )

        display_order = PayloadOrderCollator.display_order(dimension_, ())

        assert display_order == expected_value


class DescribeSortByValueCollator(object):
    """Partial-integration test suite for `SortByValueCollator` object."""

    @pytest.mark.parametrize(
        "order, xtop, xbot, element_vals, empty_idxs, expected_value",
        (
            # --- descending: subtots at top, then body ---
            ("D", [], [], (10, 30, 20, 40), (), (-2, -1, 3, 1, 2, 0)),
            ("D", [1], [], (10, 30, 20, 40), (), (-2, -1, 0, 3, 1, 2)),
            ("D", [], [4], (10, 30, 20, 40), (), (-2, -1, 1, 2, 0, 3)),
            ("D", [4], [2], (10, 30, 20, 40), (), (-2, -1, 3, 2, 0, 1)),
            ("D", [], [], (10, 30, 20, 40), (1,), (-2, -1, 3, 2, 0)),
            ("D", [3], [2], (10, 30, 20, 40), (0, 3), (-2, -1, 2, 1)),
            ("D", [3], [2], (10, 30, 20, 40), (2, 3), (-2, -1, 0, 1)),
            # --- ascending: body first, all subtots at bottom ---
            ("A", [], [], (10, 30, 20, 40), (), (0, 2, 1, 3, -1, -2)),
            ("A", [2], [], (10, 30, 20, 40), (), (1, 0, 2, 3, -1, -2)),
            ("A", [], [3], (10, 30, 20, 40), (), (0, 1, 3, 2, -1, -2)),
            ("A", [4], [1], (10, 30, 20, 40), (), (3, 2, 1, 0, -1, -2)),
            ("A", [], [], (10, 30, 20, 40), (2,), (0, 1, 3, -1, -2)),
            ("A", [4], [1], (10, 30, 20, 40), (1, 2), (3, 0, -1, -2)),
            ("A", [4], [1], (10, 30, 20, 40), (0, 3), (2, 1, -1, -2)),
        ),
    )
    def it_knows_the_display_order_for_a_dimension(
        self, order, xtop, xbot, element_vals, empty_idxs, expected_value
    ):
        subtot_vals = [60, 40]
        dimension = Dimension(
            dimension_dict={
                "type": {
                    "categories": [{"id": 1}, {"id": 2}, {"id": 3}, {"id": 4}],
                    "class": "categorical",
                }
            },
            dimension_type=DT.CAT,
            dimension_transforms={
                "order": {
                    "direction": "ascending" if order == "A" else "descending",
                    "fixed": {"top": xtop, "bottom": xbot},
                },
                "prune": True,
            },
        )

        display_order = SortByValueCollator.display_order(
            dimension, element_vals, subtot_vals, empty_idxs
        )

        assert display_order == expected_value



# encoding: utf-8

"""Integration tests for pairwise comparisons."""

from unittest import TestCase

import numpy as np
import pytest

from cr.cube.cube import Cube
from cr.cube.cubepart import _Slice

from ..fixtures import CR, OL, SM, NA
from ..util import load_python_expression


class Describe_Slice(object):
    """Integration-test suite for pairwise-t aspects of _Slice object."""

    @pytest.mark.parametrize(
        "fixture, pw_indices_dict, expectation",
        (
            (
                CR.PAIRWISE_HIROTSU_ILLNESS_X_OCCUPATION_WITH_HS,
                {},
                "cat-x-cat-hs-pw-idxs",
            ),
            (CR.PAIRWISE_HIROTSU_OCCUPATION_X_ILLNESS, {}, "cat-x-cat-pw-idxs"),
            (
                CR.PAIRWISE_HIROTSU_OCCUPATION_X_ILLNESS,
                {"only_larger": False},
                "cat-x-cat-pw-idxs-recip",
            ),
            (CR.MR_X_CAT, {}, "mr-x-cat-pw-idxs"),
            (CR.MR_X_CAT_HS_MT, {}, "mr-x-cat-hs-pw-idxs"),
            (CR.CAT_HS_X_MR, {}, "cat-hs-x-mr-pw-idxs"),
            (CR.CAT_X_MR_2, {}, "cat-x-mr-pw-idxs"),
            (CR.MR_X_MR, {}, "mr-x-mr-pw-idxs"),
        ),
    )
    def it_provides_pairwise_indices(self, fixture, pw_indices_dict, expectation):
        slice_ = _Slice(
            Cube(fixture),
            slice_idx=0,
            transforms={"pairwise_indices": pw_indices_dict},
            population=None,
            mask_size=0,
        )

        actual = slice_.pairwise_indices

        expected = np.array(load_python_expression(expectation), dtype=tuple)
        assert (expected == actual).all(), "\n%s\n\n%s" % (expected, actual)

    @pytest.mark.parametrize(
        "fixture, pw_indices_dict, expectation",
        (
            (
                CR.PAIRWISE_HIROTSU_ILLNESS_X_OCCUPATION_WITH_HS,
                {"alpha": [0.175, 0.025], "only_larger": False},
                "cat-x-cat-hs-pw-idxs-alt",
            ),
            (CR.MR_X_CAT, {"alpha": [0.175, 0.01]}, "mr-x-cat-pw-idxs-alt"),
        ),
    )
    def it_provides_pairwise_indices_alt(self, fixture, pw_indices_dict, expectation):
        """Provides indicies meeting secondary sig-test threshold, when specified."""
        slice_ = _Slice(
            Cube(fixture),
            slice_idx=0,
            transforms={"pairwise_indices": pw_indices_dict},
            population=None,
            mask_size=0,
        )

        actual = slice_.pairwise_indices_alt

        expected = np.array(load_python_expression(expectation), dtype=tuple)
        assert (expected == actual).all(), "\n%s\n\n%s" % (expected, actual)

    @pytest.mark.parametrize(
        "fixture, expectation",
        (
            (
                CR.PAIRWISE_HIROTSU_ILLNESS_X_OCCUPATION_WITH_HS,
                "cat-x-cat-hs-scale-mean-pw-idxs",
            ),
            (CR.PAIRWISE_HIROTSU_OCCUPATION_X_ILLNESS, "cat-x-cat-scale-mean-pw-idxs"),
            (CR.CAT_X_MR_2, "cat-x-mr-scale-mean-pw-idxs"),
        ),
    )
    def it_provides_columns_scale_mean_pairwise_indices(self, fixture, expectation):
        """Provides column-indicies meeting sig-test threshold on column scale means."""
        slice_ = _Slice(
            Cube(fixture), slice_idx=0, transforms={}, population=None, mask_size=0
        )

        actual = slice_.columns_scale_mean_pairwise_indices

        expected = load_python_expression(expectation)
        assert expected == actual, "\n%s\n\n%s" % (expected, actual)

    @pytest.mark.parametrize(
        "fixture, expectation",
        (
            (
                CR.PAIRWISE_HIROTSU_ILLNESS_X_OCCUPATION_WITH_HS,
                "cat-x-cat-hs-scale-mean-pw-idxs-alt",
            ),
            (
                CR.PAIRWISE_HIROTSU_OCCUPATION_X_ILLNESS,
                "cat-x-cat-scale-mean-pw-idxs-alt",
            ),
            (CR.CAT_X_MR_2, "cat-x-mr-scale-mean-pw-idxs-alt"),
        ),
    )
    def it_provides_columns_scale_mean_pairwise_indices_alt(self, fixture, expectation):
        """Provides col idxs meeting secondary sig-test threshold on scale mean."""
        slice_ = _Slice(
            Cube(fixture),
            slice_idx=0,
            transforms={"pairwise_indices": {"alpha": [0.01, 0.08]}},
            population=None,
            mask_size=0,
        )

        actual = slice_.columns_scale_mean_pairwise_indices_alt

        expected = load_python_expression(expectation)
        assert expected == actual, "\n%s\n\n%s" % (expected, actual)


class TestStandardizedResiduals(TestCase):
    """Test cr.cube implementation of column family pairwise comparisons"""

    def test_ca_subvar_hs_x_cat_hs_pairwise_t_tests(self):
        slice_ = Cube(CR.CA_SUBVAR_X_CA_CAT_HS).partitions[0]
        actual = slice_.pairwise_significance_tests[1]

        np.testing.assert_array_almost_equal(
            actual.t_stats, load_python_expression("ca-subvar-x-ca-cat-hs-pw-tstats")
        )
        np.testing.assert_array_almost_equal(
            actual.p_vals, load_python_expression("ca-subvar-x-ca-cat-hs-pw-pvals")
        )
        # check new api also
        np.testing.assert_array_almost_equal(
            slice_.pairwise_significance_t_stats(1),
            load_python_expression("ca-subvar-x-ca-cat-hs-pw-tstats"),
        )
        np.testing.assert_array_almost_equal(
            slice_.pairwise_significance_p_vals(1),
            load_python_expression("ca-subvar-x-ca-cat-hs-pw-pvals"),
        )

    def test_cat_nps_numval_x_cat_scale_means_pariwise_t_tests(self):
        slice_ = Cube(SM.CAT_NPS_NUMVAL_X_CAT).partitions[0]

        actual = slice_.pairwise_significance_tests[0]
        assert actual.p_vals_scale_means[0] == 1
        assert actual.t_stats_scale_means[0] == 0
        np.testing.assert_almost_equal(
            actual.t_stats_scale_means, [0.0, 1.08966143, -1.27412884, -4.14629088]
        )
        np.testing.assert_almost_equal(
            actual.p_vals_scale_means,
            [1.0000000e00, 2.7638510e-01, 2.0302047e-01, 4.0226310e-05],
        )

        actual = slice_.pairwise_significance_tests[1]
        assert actual.p_vals_scale_means[1] == 1
        assert actual.t_stats_scale_means[1] == 0
        np.testing.assert_almost_equal(
            actual.p_vals_scale_means,
            [2.7638510e-01, 1.0000000e00, 1.4344585e-03, 1.7905266e-09],
        )
        np.testing.assert_almost_equal(
            actual.t_stats_scale_means, [-1.08966143, 0.0, -3.19741668, -6.10466696]
        )

    def test_cat_x_cat_pairwise_t_tests(self):
        slice_ = Cube(CR.PAIRWISE_HIROTSU_OCCUPATION_X_ILLNESS).partitions[0]
        actual = slice_.pairwise_significance_tests[2]

        np.testing.assert_almost_equal(
            actual.t_stats, load_python_expression("cat-x-cat-pw-tstats")
        )
        np.testing.assert_almost_equal(
            actual.p_vals, load_python_expression("cat-x-cat-pw-pvals")
        )
        # test new api also
        np.testing.assert_almost_equal(
            slice_.pairwise_significance_t_stats(2),
            load_python_expression("cat-x-cat-pw-tstats"),
        )
        np.testing.assert_almost_equal(
            slice_.pairwise_significance_p_vals(2),
            load_python_expression("cat-x-cat-pw-pvals"),
        )

    def test_cat_x_cat_hs_pairwise_t_tests(self):
        slice_ = Cube(CR.PAIRWISE_HIROTSU_ILLNESS_X_OCCUPATION_WITH_HS).partitions[0]
        t_stats = slice_.pairwise_significance_tests[0].t_stats
        p_vals = slice_.pairwise_significance_tests[0].p_vals

        np.testing.assert_almost_equal(
            t_stats, load_python_expression("cat-x-cat-hs-pw-tstats")
        )
        np.testing.assert_almost_equal(
            p_vals, load_python_expression("cat-x-cat-hs-pw-pvals")
        )
        # test new api also
        np.testing.assert_almost_equal(
            slice_.pairwise_significance_t_stats(0),
            load_python_expression("cat-x-cat-hs-pw-tstats"),
        )
        np.testing.assert_almost_equal(
            slice_.pairwise_significance_p_vals(0),
            load_python_expression("cat-x-cat-hs-pw-pvals"),
        )

    def test_cat_x_cat_hs_scale_means_pairwise_t_tests(self):
        slice_ = Cube(CR.PAIRWISE_HIROTSU_ILLNESS_X_OCCUPATION_WITH_HS).partitions[0]
        actual = slice_.pairwise_significance_tests[0]

        np.testing.assert_almost_equal(
            actual.t_stats_scale_means,
            load_python_expression("cat-x-cat-hs-scale-means-pw-tstats"),
        )
        np.testing.assert_almost_equal(
            actual.p_vals_scale_means,
            load_python_expression("cat-x-cat-hs-scale-means-pw-pvals"),
        )

    def test_cat_x_cat_pruning_and_hs_scale_means_pairwise_t_tests(self):
        transforms = {
            "columns_dimension": {"insertions": {}},
            "rows_dimension": {"insertions": {}},
        }
        slice_ = Cube(CR.CAT_HS_MT_X_CAT_HS_MT, transforms=transforms).partitions[0]
        actual = slice_.pairwise_significance_tests[0]

        np.testing.assert_almost_equal(
            actual.t_stats_scale_means,
            [0.0, 1.64461503, 1.92387847, np.nan, 1.06912069],
        )
        np.testing.assert_almost_equal(
            actual.p_vals_scale_means, [1.0, 0.1046981, 0.059721, np.nan, 0.2918845]
        )

        # Just H&S
        slice_ = Cube(CR.CAT_HS_MT_X_CAT_HS_MT).partitions[0]
        actual = slice_.pairwise_significance_tests[0]

        np.testing.assert_almost_equal(
            actual.t_stats_scale_means,
            [0.0, 0.9387958, 1.644615, 1.9238785, np.nan, 1.0691207],
        )

        np.testing.assert_almost_equal(
            actual.p_vals_scale_means,
            [1.0, 0.3500141, 0.1046981, 0.059721, np.nan, 0.2918845],
        )

        # Just pruning
        transforms = {
            "rows_dimension": {"prune": True},
            "columns_dimension": {"prune": True},
        }
        slice_ = Cube(CR.CAT_HS_MT_X_CAT_HS_MT, transforms=transforms).partitions[0]
        actual = slice_.pairwise_significance_tests[0]

        np.testing.assert_almost_equal(
            actual.t_stats_scale_means,
            [0.0, 0.93879579, 1.64461503, 1.92387847, 1.06912069],
        )
        np.testing.assert_almost_equal(
            actual.p_vals_scale_means, [1.0, 0.3500141, 0.1046981, 0.059721, 0.2918845]
        )

        # Pruning and H&S
        transforms = {
            "rows_dimension": {"insertions": {}, "prune": True},
            "columns_dimension": {"insertions": {}, "prune": True},
        }
        slice_ = Cube(CR.CAT_HS_MT_X_CAT_HS_MT, transforms=transforms).partitions[0]
        actual = slice_.pairwise_significance_tests[0]

        np.testing.assert_almost_equal(
            actual.t_stats_scale_means, [0.0, 1.64461503, 1.92387847, 1.06912069]
        )
        np.testing.assert_almost_equal(
            actual.p_vals_scale_means, [1.0, 0.1046981, 0.059721, 0.2918845]
        )

    def test_cat_x_cat_summary_pairwise_indices(self):
        # Only larger
        slice_ = Cube(CR.PAIRWISE_HIROTSU_OCCUPATION_X_ILLNESS).partitions[0]
        pairwise_indices = slice_.summary_pairwise_indices
        expected_indices = np.array([(2,), (0, 2), ()], dtype=tuple)
        np.testing.assert_array_equal(pairwise_indices, expected_indices)

        # Larger and smaller
        transforms = {"pairwise_indices": {"only_larger": False}}
        slice_ = Cube(
            CR.PAIRWISE_HIROTSU_OCCUPATION_X_ILLNESS, transforms=transforms
        ).partitions[0]
        pairwise_indices = slice_.summary_pairwise_indices
        expected_indices = np.array([(1, 2), (0, 2), (0, 1)], dtype="i,i")
        np.testing.assert_array_equal(pairwise_indices, expected_indices)

    def test_cat_x_cat_wgtd_pairwise_t_tests(self):
        """The weights on this cube demonstrate much higher variance (less
        extreme t values, and higher associated p-values) than if weighted_n
        were used in the variance estimate of the test statistic.
        """
        slice_ = Cube(CR.CAT_X_CAT_WEIGHTED_TTESTS).partitions[0]
        actual = slice_.pairwise_significance_tests[0]
        pairwise_indices = slice_.pairwise_indices

        np.testing.assert_almost_equal(
            actual.t_stats, load_python_expression("cat-x-cat-wgtd-pw-tstats")
        )
        np.testing.assert_almost_equal(
            actual.p_vals, load_python_expression("cat-x-cat-wgtd-pw-pvals")
        )
        np.testing.assert_array_equal(
            pairwise_indices,
            np.array(load_python_expression("cat-x-cat-wgtd-pw-indices"), dtype=tuple),
        )
        # test new api also
        np.testing.assert_almost_equal(
            slice_.pairwise_significance_t_stats(0),
            load_python_expression("cat-x-cat-wgtd-pw-tstats"),
        )
        np.testing.assert_almost_equal(
            slice_.pairwise_significance_p_vals(0),
            load_python_expression("cat-x-cat-wgtd-pw-pvals"),
        )

    def test_cat_x_cat_wgtd_scale_means_pariwise_t_tests(self):
        slice_ = Cube(CR.CAT_X_CAT_WEIGHTED_TTESTS).partitions[0]
        actual = slice_.pairwise_significance_tests[0]

        np.testing.assert_almost_equal(
            actual.t_stats_scale_means, [0.0, -4.38871748, -3.99008596, -5.15679647]
        )
        np.testing.assert_almost_equal(
            actual.p_vals_scale_means,
            [1.0000000e00, 1.3839564e-05, 7.4552516e-05, 4.0145665e-07],
        )

    def test_mr_x_mr_pairwise_t_tests(self):
        slice_ = Cube(CR.MR_X_MR_2).partitions[0]
        actual = slice_.pairwise_significance_tests[1]

        np.testing.assert_array_almost_equal(
            actual.t_stats, load_python_expression("mr-x-mr-pw-tstats")
        )
        np.testing.assert_array_almost_equal(
            actual.p_vals, load_python_expression("mr-x-mr-pw-pvals")
        )
        # test new api also
        np.testing.assert_array_almost_equal(
            slice_.pairwise_significance_t_stats(1),
            load_python_expression("mr-x-mr-pw-tstats"),
        )
        np.testing.assert_array_almost_equal(
            slice_.pairwise_significance_p_vals(1),
            load_python_expression("mr-x-mr-pw-pvals"),
        )


class TestOverlapsPairwiseSignificance(TestCase):
    def test_pairwise_significance_cat_x_mr_sub_x_mr_sel_0th_subvar(self):
        slice_ = Cube(OL.CAT_X_MR_SUB_X_MR_SEL).partitions[0]

        assert slice_.column_percentages.tolist() == [
            [0.0, 100.0, 100.0],
            [100.0, 0.0, 0.0],
        ]
        assert slice_.pairwise_significance_t_stats(0) == pytest.approx(
            np.array(
                [
                    [0.0, 3.11420549, 2.61911361],
                    [0.0, -3.11420549, -2.61911361],
                ]
            )
        )
        assert slice_.pairwise_significance_p_vals(0) == pytest.approx(
            np.array(
                [
                    [0.0, 0.05270861, 0.07906174],
                    [0.0, 0.05270861, 0.07906174],
                ]
            )
        )

    def test_pairwise_significance_cat_x_mr_sub_x_mr_sel_1st_subvar(self):
        slice_ = Cube(OL.CAT_X_MR_SUB_X_MR_SEL).partitions[0]

        assert slice_.column_percentages.tolist() == [
            [0.0, 100.0, 100.0],
            [100.0, 0.0, 0.0],
        ]
        assert slice_.pairwise_significance_t_stats(1) == pytest.approx(
            np.array(
                [
                    [-3.11420549, 0.0, 0.0],
                    [3.11420549, 0.0, 0.0],
                ]
            )
        )
        assert slice_.pairwise_significance_p_vals(1) == pytest.approx(
            np.array(
                [
                    [0.05270861, 0.0, 1.0],
                    [0.05270861, 0.0, 1.0],
                ]
            )
        )

    def test_pairwise_significance_cat_x_mr_sub_x_mr_sel_2nd_subvar(self):
        slice_ = Cube(OL.CAT_X_MR_SUB_X_MR_SEL).partitions[0]

        assert slice_.column_percentages.tolist() == [
            [0.0, 100.0, 100.0],
            [100.0, 0.0, 0.0],
        ]
        assert slice_.pairwise_significance_t_stats(2) == pytest.approx(
            np.array(
                [
                    [-2.61911361, 0.0, 0.0],
                    [2.61911361, 0.0, 0.0],
                ],
            ),
        )
        assert slice_.pairwise_significance_p_vals(2) == pytest.approx(
            np.array(
                [
                    [0.07906174, 1.0, 0.0],
                    [0.07906174, 1.0, 0.0],
                ],
            ),
        )

    def test_pairwise_significance_cat_x_mr_realistic_example(self):
        slice_ = Cube(OL.CAT_X_MR_REALISTIC_EXAMPLE).partitions[0]

        assert slice_.column_percentages == pytest.approx(
            np.array(
                [
                    [52.7687, 52.5926, 51.5504, 47.6852, 51.3889, np.nan],
                    [47.2313, 47.4074, 48.4496, 52.3148, 48.6111, np.nan],
                ]
            ),
            nan_ok=True,
        )
        assert slice_.pairwise_significance_t_stats(4) == pytest.approx(
            np.array(
                [
                    [1.00337549, 0.64382181, 0.0773666, -1.3677023, 0.0, np.nan],
                    [-1.00337549, -0.64382181, -0.0773666, 1.3677023, 0.0, np.nan],
                ],
            ),
            nan_ok=True,
        )
        assert slice_.pairwise_significance_p_vals(4) == pytest.approx(
            np.array(
                [
                    [0.31647509, 0.52017481, 0.93838264, 0.17241219, 0.0, np.nan],
                    [0.31647509, 0.52017481, 0.93838264, 0.17241219, 0.0, np.nan],
                ],
            ),
            nan_ok=True,
        )

    def test_pairwise_significance_mr_x_mr(self):
        slice_ = Cube(OL.MR_X_MR).partitions[0]

        assert slice_.column_percentages == pytest.approx(
            np.array(
                [
                    [100.0, 66.66667, np.nan],
                    [66.66667, 100.0, np.nan],
                    [0.0, 0.0, np.nan],
                ]
            ),
            nan_ok=True,
        )
        assert slice_.pairwise_significance_t_stats(1) == pytest.approx(
            np.array(
                [
                    [1.22474487, 0.0, np.nan],
                    [-1.22474487, np.nan, np.nan],
                    [np.nan, np.nan, np.nan],
                ],
            ),
            nan_ok=True,
        )
        assert slice_.pairwise_significance_p_vals(1) == pytest.approx(
            np.array(
                [
                    [0.28786413, 1.0, np.nan],
                    [0.27521973, np.nan, np.nan],
                    [np.nan, np.nan, np.nan],
                ],
            ),
            nan_ok=True,
        )

    def test_pairwise_cat_x_mr_gender_x_all_pets_owned_with_weighted_counts(self):
        slice_ = Cube(OL.CAT_X_MR_GENDER_X_ALL_PETS_OWNED).partitions[0]

        assert slice_.column_percentages.tolist() == pytest.approx(
            np.array(
                [
                    [66.6667, 14.28571, 50.0],
                    [33.33333, 85.714286, 50.0],
                ]
            )
        )

        # Assert for first column (subvariable)
        assert slice_.pairwise_significance_t_stats(0).tolist() == pytest.approx(
            np.array(
                [
                    [0.0, -2.6315597, -1.76353],
                    [0.0, 2.6315597, 1.76353],
                ]
            ),
        )
        assert slice_.pairwise_significance_p_vals(0) == pytest.approx(
            np.array(
                [
                    [0.0, 0.01410448, 0.0879948],
                    [0.0, 0.01410448, 0.0879948],
                ]
            ),
        )

        # Assert for second column (subvariable)
        assert slice_.pairwise_significance_t_stats(1).tolist() == pytest.approx(
            np.array(
                [
                    [2.63156, 0.0, 8.10444],
                    [-2.63156, 0.0, -8.10444],
                ]
            ),
        )
        assert slice_.pairwise_significance_p_vals(1) == pytest.approx(
            np.array(
                [
                    [0.01410448, 0, 0.025067e-06],
                    [0.01410448, 0, 0.025067e-06],
                ]
            ),
        )

        # Assert for third column (subvariable)
        assert slice_.pairwise_significance_t_stats(2).tolist() == pytest.approx(
            np.array(
                [
                    [1.763531, -8.104439, 0.0],
                    [-1.763531, 8.104439, 0.0],
                ]
            ),
        )
        assert slice_.pairwise_significance_p_vals(2) == pytest.approx(
            np.array(
                [
                    [0.0879948, 0.025067e-06, 0],
                    [0.0879948, 0.025067e-06, 0],
                ]
            ),
        )

    def test_pairwise_significance_indices(self):
        transforms = {"pairwise_indices": {"alpha": [0.05, 0.13]}}
        slice_ = Cube(
            OL.CAT_X_MR_GENDER_X_ALL_PETS_OWNED, transforms=transforms
        ).partitions[0]

        assert slice_.column_percentages.tolist() == pytest.approx(
            np.array(
                [
                    [66.6667, 14.28571, 50.0],
                    [33.33333, 85.714286, 50.0],
                ]
            )
        )

        assert slice_.pairwise_indices.tolist() == [
            [(1,), (), (1,)],
            [(), (0, 2), ()],
        ]

        assert slice_.pairwise_indices_alt.tolist() == [
            [(1, 2), (), (1,)],
            [(), (0, 2), (0,)],
        ]

    def test_alt_pairwise_indices_without_alt_alpha(self):
        slice_ = Cube(OL.CAT_X_MR_GENDER_X_ALL_PETS_OWNED).partitions[0]
        assert slice_.pairwise_indices_alt is None

    def test_pairwise_significance_all_empty(self):
        # ---Keep the alpha value this small to demo the error found in cr.server
        transforms = {"pairwise_indices": {"alpha": [0.0000000001]}}
        slice_ = Cube(
            OL.CAT_X_MR_GENDER_X_ALL_PETS_OWNED, transforms=transforms
        ).partitions[0]

        assert slice_.column_percentages.tolist() == pytest.approx(
            np.array(
                [
                    [66.6667, 14.28571, 50.0],
                    [33.33333, 85.714286, 50.0],
                ]
            )
        )

        assert slice_.pairwise_indices.tolist() == [
            [(), (), ()],
            [(), (), ()],
        ]

    def test_pairwise_sig_for_realistic_example_mr_x_mr(self):
        slice_ = Cube(OL.MR_X_MR_REALISTIC_EXAMPLE).partitions[0]
        assert slice_.pairwise_significance_t_stats(3) == pytest.approx(
            np.array(
                [
                    [13.40, 9.62, -0.49, 0.0, 2.27, -24.08],
                    [-0.58, 19.94, -4.76, 0.0, -2.88, -13.48],
                    [-0.53, 0.70, 22.76, 0.0, 10.32, -19.63],
                    [-18.79, -13.57, -20.00, 0.0, -13.14, -29.10],
                    [-5.45, -0.99, -2.83, 0.0, 21.75, -12.29],
                    [-0.76, 0.17, -3.35, 0.0, -1.07, 31.68],
                ]
            ),
            abs=10e-2,
        )

    def test_pairwise_sig_for_mr_x_mr_vs_mr_single_subvar_x_mr(self):
        mr_x_mr_slice = Cube(OL.MR_X_MR).partitions[0]
        t_stats_mr_x_mr = mr_x_mr_slice.pairwise_significance_t_stats(1)
        mr_subvar_x_mr_slice = Cube(OL.MR_SINGLE_SUBVAR_X_MR).partitions[0]
        t_stats_mr_subvar_x_mr = mr_subvar_x_mr_slice.pairwise_significance_t_stats(1)
        # Assert same row stats are the same in both cases (MR x MR and MR_SEL x MR)
        np.testing.assert_array_equal(t_stats_mr_x_mr[0], t_stats_mr_subvar_x_mr[0])


class TestMeanDifferenceSignificance(object):
    def test_mean_diff_significance_for_numeric_array_grouped_by_cat(self):
        slice_ = Cube(NA.NUM_ARR_MULTI_NUMERIC_MEASURES_GROUPED_BY_CAT).partitions[0]
        assert slice_.pairwise_significance_means_t_stats(0) == pytest.approx(
            np.array(
                [
                    [-0.0, -0.7482866, -4.61524529, -5.29463547],
                    [-0.0, -4.64371041, -7.14738681, -3.67511747],
                    [-0.0, 3.13387218, 6.35310465, 2.3932706],
                ]
            )
        )
        assert slice_.pairwise_significance_means_p_vals(0) == pytest.approx(
            np.array(
                [
                    [1.00000000e00, 4.60526868e-01, 5.38716457e-05, 7.11866628e-06],
                    [1.00000000e00, 7.34317022e-05, 2.90876736e-08, 8.12768987e-04],
                    [1.00000000e00, 4.02303806e-03, 3.00760739e-07, 2.23647892e-02],
                ]
            )
        )

    def test_mean_diff_significance_for_numeric_array_grouped_by_cat_hs(self):
        transforms = {
            "columns_dimension": {
                "insertions": [
                    {
                        "function": "subtotal",
                        "name": "DIFF B-A",
                        "args": [4, 5],
                        "anchor": "top",
                        "kwargs": {"negative": [1, 2, 3]},
                    },
                    {
                        "function": "subtotal",
                        "args": [1, 2, 3],
                        "anchor": 1,
                        "name": '"A" countries',
                    },
                    {
                        "function": "subtotal",
                        "args": [4, 5],
                        "anchor": 2,
                        "name": '"B" countries',
                    },
                    {
                        "function": "subtotal",
                        "name": "DIFF A-B",
                        "args": [1, 2, 3],
                        "anchor": "bottom",
                        "kwargs": {"negative": [4, 5]},
                    },
                ]
            }
        }
        slice_ = Cube(
            NA.NUM_ARR_MEANS_GROUPED_BY_CAT_HS, transforms=transforms
        ).partitions[0]

        # Column (0,3,5,8) are subtotals and the hypothesis testing for a subtotal
        # computable ATM.
        for col in slice_.inserted_column_idxs:
            assert slice_.pairwise_significance_means_t_stats(
                col
            ).tolist() == pytest.approx(
                np.full(slice_.means.shape, np.nan), nan_ok=True
            )
        np.testing.assert_almost_equal(
            slice_.pairwise_significance_means_t_stats(7),
            load_python_expression("num-arr-means-grouped-by-cat-hs-t-stats-col-7"),
        )
        np.testing.assert_almost_equal(
            slice_.pairwise_significance_means_p_vals(7),
            load_python_expression("num-arr-means-grouped-by-cat-hs-p-vals-col-7"),
        )
        assert slice_.pairwise_means_indices.tolist() == [
            [None, (), (7,), None, (), None, (), (), None],
            [None, (), (), None, (), None, (), (), None],
            [None, (), (1, 4, 7), None, (), None, (1, 4, 7), (), None],
            [None, (), (), None, (), None, (), (), None],
        ]

        # Test no subtotals
        slice_ = Cube(NA.NUM_ARR_MEANS_GROUPED_BY_CAT_HS).partitions[0]
        assert slice_.pairwise_means_indices.tolist() == [
            [(), (4,), (), (), ()],
            [(), (), (), (), ()],
            [(), (0, 2, 4), (), (0, 2, 4), ()],
            [(), (), (), (), ()],
        ]
        np.testing.assert_almost_equal(
            slice_.pairwise_significance_means_t_stats(4),
            load_python_expression("num-arr-means-grouped-by-cat-t-stats-col-4"),
        )
        np.testing.assert_almost_equal(
            slice_.pairwise_significance_means_p_vals(4),
            load_python_expression("num-arr-means-grouped-by-cat-p-vals-col-4"),
        )

    def test_mean_diff_significance_for_numeric_array_x_mr(self):
        slice_ = Cube(NA.NUM_ARR_MULTI_NUMERIC_MEASURES_X_MR).partitions[0]

        assert slice_.pairwise_significance_means_t_stats(0) == pytest.approx(
            np.array(
                [
                    [-0.0, 1.3056921, -1.1842639],
                    [-0.0, 1.7117695, -1.6297401],
                    [np.nan, np.nan, np.nan],
                    [np.nan, -5.8872721, -4.4442801],
                ]
            ),
            nan_ok=True,
        )
        assert slice_.pairwise_significance_means_p_vals(0) == pytest.approx(
            np.array(
                [
                    [1.00000000e00, 1.96331347e-01, 2.40688158e-01],
                    [1.00000000e00, 9.17815019e-02, 1.08068956e-01],
                    [np.nan, np.nan, np.nan],
                    [np.nan, 1.57810759e-07, 3.57756031e-05],
                ]
            ),
            nan_ok=True,
        )

    def test_mean_diff_significance_indices_for_numeric_array_grouped_by_cat(self):
        transforms = {"pairwise_indices": {"alpha": [0.05, 0.08]}}
        slice_ = Cube(
            NA.NUM_ARR_MULTI_NUMERIC_MEASURES_GROUPED_BY_CAT, transforms=transforms
        ).partitions[0]

        assert slice_.pairwise_means_indices.tolist() == [
            [(2, 3), (2, 3), (), ()],
            [(1, 2, 3), (), (), ()],
            [(), (0,), (0, 1, 3), (0,)],
        ]
        assert slice_.pairwise_means_indices_alt.tolist() == [
            [(2, 3), (2, 3), (), ()],
            [(1, 2, 3), (), (), (2,)],
            [(), (0,), (0, 1, 3), (0,)],
        ]

    def test_mean_diff_significance_is_not_available(self):
        transforms = {"pairwise_indices": {"alpha": [0.05, 0.08]}}
        slice_ = Cube(CR.CAT_X_CAT, transforms=transforms).partitions[0]

        with pytest.raises(ValueError) as e:
            slice_.pairwise_means_indices
        assert (
            str(e.value) == "`.pairwise_means_indices` is undefined for a cube-result"
            " without a mean measure"
        )
        with pytest.raises(ValueError) as e:
            slice_.pairwise_means_indices_alt
        assert (
            str(e.value)
            == "`.pairwise_means_indices_alt` is undefined for a cube-result"
            " without a mean measure"
        )
        with pytest.raises(ValueError) as e:
            slice_.pairwise_significance_means_p_vals(0)
        assert (
            str(e.value)
            == "`.pairwise_significance_means_p_vals` is undefined for a cube-result"
            " without a mean measure"
        )
        with pytest.raises(ValueError) as e:
            slice_.pairwise_significance_means_t_stats(0)
        assert (
            str(e.value)
            == "`.pairwise_significance_means_t_stats` is undefined for a cube-result"
            " without a mean measure"
        )

    def test_mean_diff_significance_for_cat_x_cat(self):
        slice_ = Cube(CR.MEAN_CAT_X_CAT).partitions[0]

        assert slice_.pairwise_significance_means_t_stats(1) == pytest.approx(
            np.array(
                [
                    [np.nan, np.nan, np.nan, np.nan],
                    [14.1421356, -0.0, 3.1066887, np.nan],
                    [1.75662013, -0.0, np.nan, np.nan],
                ]
            ),
            nan_ok=True,
        )
        assert slice_.pairwise_significance_means_p_vals(1) == pytest.approx(
            np.array(
                [
                    [np.nan, np.nan, np.nan, np.nan],
                    [8.21565038e-15, 1.00000000e00, 4.80968355e-03, np.nan],
                    [8.91869315e-02, 1.00000000e00, np.nan, np.nan],
                ]
            ),
            nan_ok=True,
        )

    def test_mean_diff_significance_indices_for_cat_x_cat(self):
        transforms = {"pairwise_indices": {"alpha": [0.05, 0.01]}}
        slice_ = Cube(CR.MEAN_CAT_X_CAT, transforms=transforms).partitions[0]

        assert slice_.pairwise_means_indices.tolist() == [
            [(), (), (), ()],
            [(1, 2), (), (1,), ()],
            [(), (), (), ()],
        ]
        assert slice_.pairwise_means_indices_alt.tolist() == [
            [(), (), (), (0, 2)],
            [(1, 2), (), (1,), ()],
            [(), (), (), ()],
        ]

    def test_mean_diff_significance_indices_num_array_grouped_by_cat_hs_weighted(self):
        insertions = [
            {
                "function": "subtotal",
                "args": [1, 2, 3],
                "anchor": "top",
                "name": '"A" countries',
            },
            {
                "function": "subtotal",
                "args": [4, 5],
                "anchor": "top",
                "name": '"B" countries',
            },
            {
                "function": "subtotal",
                "name": "DIFF A-B",
                "args": [1, 2, 3],
                "anchor": "top",
                "kwargs": {"negative": [4, 5]},
            },
            {
                "function": "subtotal",
                "name": "DIFF B-A",
                "args": [4, 5],
                "anchor": "top",
                "kwargs": {"negative": [1, 2, 3]},
            },
        ]
        transforms = {
            "pairwise_indices": {"alpha": [0.05, 0.01]},
            "columns_dimension": {"insertions": insertions},
        }
        slice_ = Cube(
            NA.NUM_ARR_MEANS_GROUPED_BY_CAT_HS_WEIGHTED, transforms=transforms
        ).partitions[0]
        assert slice_.pairwise_means_indices.tolist() == [
            [None, None, None, None, (), (), (), (), ()],
            [None, None, None, None, (), (), (), (), ()],
            [None, None, None, None, (), (), (), (6,), ()],
        ]
        assert slice_.pairwise_means_indices_alt.tolist() == [
            [None, None, None, None, (), (), (), (), ()],
            [None, None, None, None, (), (), (), (), ()],
            [None, None, None, None, (), (), (), (4, 6), (4, 6)],
        ]

        # Test no subtotals
        transforms = {"pairwise_indices": {"alpha": [0.05, 0.01]}}
        slice_ = Cube(
            NA.NUM_ARR_MEANS_GROUPED_BY_CAT_HS_WEIGHTED, transforms=transforms
        ).partitions[0]
        assert slice_.pairwise_means_indices.tolist() == [
            [(), (), (), (), ()],
            [(), (), (), (), ()],
            [(), (), (), (2,), ()],
        ]
        assert slice_.pairwise_means_indices_alt.tolist() == [
            [(), (), (), (), ()],
            [(), (), (), (), ()],
            [(), (), (), (0, 2), (0, 2)],
        ]

        # Test pruning
        transforms = {
            "columns_dimension": {
                "prune": True,
                "elements": {"1": {"hide": True}},
                "insertions": insertions,
            },
            "pairwise_indices": {"alpha": [0.05, 0.01]},
        }
        slice_ = Cube(
            NA.NUM_ARR_MEANS_GROUPED_BY_CAT_HS_WEIGHTED, transforms=transforms
        ).partitions[0]

        assert slice_.pairwise_means_indices.tolist() == [
            [None, None, None, None, (), (), (), ()],
            [None, None, None, None, (), (), (), ()],
            [None, None, None, None, (), (), (5,), ()],
        ]
        assert slice_.pairwise_means_indices_alt.tolist() == [
            [None, None, None, None, (), (), (), ()],
            [None, None, None, None, (), (), (), ()],
            [None, None, None, None, (), (), (4, 5), (4, 5)],
        ]

    def test_mean_indices_cat_x_mr_subvar(self):
        slice_ = Cube(CR.CAT_X_MR_SUBVAR_MEANS).partitions[0]

        assert slice_.pairwise_means_indices == pytest.approx(
            np.array([[(), (), (), ()]])
        )



# encoding: utf-8

"""Provides subtotalling services according to a variety of strategies.

A matrix can have inserted vectors (row or column subtotals) that summarize two or more
other vectors by ostensibly "adding" them. Simple addition works for counts, but more
sophisticated methods are required for higher-order measures.

This module provides the various strategies required for computing subtotals and is
primarily used by measure objects as a collaborator to handle this aspect.
"""

from __future__ import division

import numpy as np

from cr.cube.util import lazyproperty


class _BaseSubtotals(object):
    """Base class for Subtotals objects."""

    def __init__(self, base_values, dimensions):
        self._base_values = base_values
        self._dimensions = dimensions

    @classmethod
    def blocks(cls, base_values, dimensions):
        """Return base, row and col insertion, and intersection matrices.

        These are in the form ready for assembly.
        """
        return cls(base_values, dimensions)._blocks

    @lazyproperty
    def _blocks(self):
        """base, row and col insertion, and intersection matrices."""
        return [
            [self._base_values, self._subtotal_columns],
            [self._subtotal_rows, self._intersections],
        ]

    @lazyproperty
    def _column_subtotals(self):
        """Sequence of _Subtotal object for each subtotal in columns-dimension."""
        return self._columns_dimension.subtotals

    @lazyproperty
    def _columns_dimension(self):
        """Dimension object for matrix columns."""
        return self._dimensions[1]

    def _intersection(self, row_subtotal, column_subtotal):
        """Value for this row/column subtotal intersection."""
        raise NotImplementedError(
            "`%s` must implement `._intersection()`" % type(self).__name__
        )  # pragma: no cover

    @lazyproperty
    def _intersections(self):
        """(n_row_subtotals, n_col_subtotals) ndarray of intersection values.

        An intersection value arises where a row-subtotal crosses a column-subtotal.
        """
        return np.array(
            [
                self._intersection(row_subtotal, column_subtotal)
                for row_subtotal in self._row_subtotals
                for column_subtotal in self._column_subtotals
            ],
        ).reshape(len(self._row_subtotals), len(self._column_subtotals))

    @lazyproperty
    def _ncols(self):
        """int count of columns in base-matrix."""
        return self._base_values.shape[1]

    @lazyproperty
    def _nrows(self):
        """int count of rows in base-matrix."""
        return self._base_values.shape[0]

    @lazyproperty
    def _row_subtotals(self):
        """Sequence of _Subtotal object for each subtotal in rows-dimension."""
        return self._rows_dimension.subtotals

    @lazyproperty
    def _rows_dimension(self):
        """Dimension object for rows of matrix."""
        return self._dimensions[0]

    def _subtotal_column(self, subtotal):
        """Return (n_rows,) ndarray of values for `subtotal` column."""
        raise NotImplementedError(
            "`%s` must implement `._subtotal_column()`" % type(self).__name__
        )  # pragma: no cover

    @lazyproperty
    def _subtotal_columns(self):
        """(n_rows, n_col_subtotals) matrix of subtotal columns."""
        subtotals = self._column_subtotals

        if len(subtotals) == 0:
            return np.empty((self._nrows, 0))

        return np.hstack(
            [
                self._subtotal_column(subtotal).reshape(self._nrows, 1)
                for subtotal in subtotals
            ]
        )

    def _subtotal_row(self, subtotal):
        """Return (n_cols,) ndarray of values for `subtotal` row."""
        raise NotImplementedError(
            "`%s` must implement `._subtotal_row()`" % type(self).__name__
        )  # pragma: no cover

    @lazyproperty
    def _subtotal_rows(self):
        """(n_row_subtotals, n_cols) ndarray of subtotal rows."""
        subtotals = self._row_subtotals

        if len(subtotals) == 0:
            return np.empty((0, self._ncols))

        return np.vstack(
            [self._subtotal_row(subtotal) for subtotal in self._row_subtotals]
        )


class NanSubtotals(_BaseSubtotals):
    """Subtotal blocks for measures that cannot meaningfully be subtotaled.

    Each subtotal value (and intersection value) is `np.nan`.
    """

    filler = np.nan

    def _intersection(self, row_subtotal, column_subtotal):
        """Unconditionally return np.nan for each intersection cell."""
        return self.filler

    def _subtotal_column(self, subtotal):
        """Return (n_rows,) ndarray of np.nan values."""
        return np.full(self._nrows, self.filler)

    def _subtotal_row(self, subtotal):
        """Return (n_cols,) ndarray of np.nan values."""
        return np.full(self._ncols, self.filler)


class NoneSubtotals(NanSubtotals):
    filler = None


class SumSubtotals(_BaseSubtotals):
    """Subtotal "blocks" created by adding and subtracting terms for subtotals.

    In addition to `base_values` and `dimensions`, `SumSubtotals` have
    properties for `diff_cols_nan` and `diff_rows_nan` which allow for columns/
    rows where subtotals have subtrahends to be overridden with np.nan. These are
    used for measures such as bases (and measures derived from bases, like proportions)
    that are not computed along the same direction as a subtotal difference. Example:
    a subtotal difference in the row dimension does not have a valid row proportion.

    One way of thinking about this is that when calculating proportions, users are
    only interested in the direction of proportions where the difference is equal
    to the sum of the addends minus the sum of the subtrahends. But when calculating
    proportions along a row or column, the proportions are only additive in one
    direction, so subtotal differences in the other direction don't work. For example,
    when you go along a row and you are calculating column percents, each column
    has a different base, so the proportions don't add up.

    Example showing column percents. We do want to calculate column proportion for
    `a-b`, but not `c-d` because `c-d` is a subtotal difference in the column
    dimension.
    ```
    |     | c-d | c   | d   | c+d  |
    |-----|-----|-----|-----|------|
    | a-b | -   | 20  | 50  | 25   |
    | a   | -   | 60  | 75  | 62.5 |
    | b   | -   | 40  | 25  | 37.5 |
    | a+b | -   | 100 | 100 | 100  |
    ```

    Another way to think about it is that a "base" is a positive thing, so you can't
    really subtract out the subtrahends. One option would be to add both the addends and
    the subtrahends, but ultimately we decided that users would be confused by
    by these values, and they really only want to see the base and proportions in the
    opposing direction. Therefore we set the corresponding direction to nan.
    """

    def __init__(
        self, base_values, dimensions, diff_cols_nan=False, diff_rows_nan=False
    ):
        super(SumSubtotals, self).__init__(base_values, dimensions)
        self._diff_cols_nan = diff_cols_nan
        self._diff_rows_nan = diff_rows_nan

    @classmethod
    def blocks(cls, base_values, dimensions, diff_cols_nan=False, diff_rows_nan=False):
        """Return base, row and col insertion, and intersection matrices.

        These are in the form ready for assembly.

        Keyword arguments:
        `diff_cols_nan` -- Overrides subtotal differences in the columns direction eg for
        column bases (default False)
        `diff_rows_nan` -- Overrides subtotal differences in the rows direction eg for
        row bases (default False)
        """
        return cls(base_values, dimensions, diff_cols_nan, diff_rows_nan)._blocks

    @classmethod
    def intersections(
        cls, base_values, dimensions, diff_cols_nan=False, diff_rows_nan=False
    ):
        """Return (n_row_subtotals, n_col_subtotals) ndarray of intersection values.

        An intersection value arises where a row-subtotal crosses a column-subtotal.

        Keyword arguments:
        `diff_cols_nan` -- Overrides subtotal differences in the columns direction eg for
        column bases (default False)
        `diff_rows_nan` -- Overrides subtotal differences in the rows direction eg for
        row bases (default False)
        """
        return cls(base_values, dimensions, diff_cols_nan, diff_rows_nan)._intersections

    @classmethod
    def subtotal_columns(
        cls, base_values, dimensions, diff_cols_nan=False, diff_rows_nan=False
    ):
        """Return (n_base_rows, n_col_subtotals) ndarray of subtotal columns.

        Keyword arguments:
        `diff_cols_nan` -- Overrides subtotal differences in the columns direction eg for
        column bases (default False)
        `diff_rows_nan` -- Overrides subtotal differences in the rows direction eg for
        row bases (default False)
        """
        return cls(
            base_values, dimensions, diff_cols_nan, diff_rows_nan
        )._subtotal_columns

    @classmethod
    def subtotal_rows(
        cls, base_values, dimensions, diff_cols_nan=False, diff_rows_nan=False
    ):
        """Return (n_row_subtotals, n_base_cols) ndarray of subtotal rows.

        Keyword arguments:
        `diff_cols_nan` -- Overrides subtotal differences in the columns direction eg for
        column bases (default False)
        `diff_rows_nan` -- Overrides subtotal differences in the rows direction eg for
        row bases (default False)
        """
        return cls(base_values, dimensions, diff_cols_nan, diff_rows_nan)._subtotal_rows

    def _intersection(self, row_subtotal, column_subtotal):
        """Sum and Diff for this row/column subtotal intersection."""

        col_has_subs = len(column_subtotal.subtrahend_idxs) > 0
        row_has_subs = len(row_subtotal.subtrahend_idxs) > 0

        if (
            # --- Intersections of subtotal differences are undefined ---
            (col_has_subs and row_has_subs)
            # --- Also need to respect diff_cols_nan/diff_rows_nan
            or (col_has_subs and self._diff_cols_nan)
            or (row_has_subs and self._diff_rows_nan)
        ):
            return np.nan

        addend_sum = np.sum(
            self._subtotal_row(row_subtotal)[column_subtotal.addend_idxs]
        )
        subtrahend_sum = np.sum(
            self._subtotal_row(row_subtotal)[column_subtotal.subtrahend_idxs]
        )
        return addend_sum - subtrahend_sum

    def _subtotal_column(self, subtotal):
        """Return (n_rows,) ndarray of values for `subtotal` column."""
        if self._diff_cols_nan and len(subtotal.subtrahend_idxs) > 0:
            return np.full(self._base_values.shape[0], np.nan)

        addend_sum = np.sum(self._base_values[:, subtotal.addend_idxs], axis=1)
        subtrahend_sum = np.sum(self._base_values[:, subtotal.subtrahend_idxs], axis=1)
        return addend_sum - subtrahend_sum

    def _subtotal_row(self, subtotal):
        """Return (n_cols,) ndarray of values for `subtotal` row."""
        if self._diff_rows_nan and len(subtotal.subtrahend_idxs) > 0:
            return np.full(self._base_values.shape[1], np.nan)

        addend_sum = np.sum(self._base_values[subtotal.addend_idxs, :], axis=0)
        subtrahend_sum = np.sum(self._base_values[subtotal.subtrahend_idxs, :], axis=0)
        return addend_sum - subtrahend_sum


class TableStdErrSubtotals(_BaseSubtotals):
    """Computes subtotal values for the table-stderrs measure."""

    def __init__(self, base_values, dimensions, cube_result_matrix):
        super(TableStdErrSubtotals, self).__init__(base_values, dimensions)
        self._cube_result_matrix = cube_result_matrix

    @classmethod
    def blocks(cls, base_values, dimensions, cube_result_matrix):
        """Return base, row and col insertion, and intersection matrices.

        These are in the form ready for assembly.
        """
        return cls(base_values, dimensions, cube_result_matrix)._blocks

    @lazyproperty
    def _base_counts(self):
        """2D np.float64 ndarray of weighted-count for each cell of base matrix."""
        return self._cube_result_matrix.weighted_counts

    def _intersection(self, row_subtotal, column_subtotal):
        """Return value for intersection of `row_subtotal` and `column_subtotal`."""
        # --- Cannot calculate if there are any subtrahends ---
        if (
            len(row_subtotal.subtrahend_idxs) > 0
            or len(column_subtotal.subtrahend_idxs) > 0
        ):
            return np.nan

        # --- column of base subtotal counts is 1D, like [44 148 283 72] ---
        # --- the inserted-row for counts measure, like: [159 172 107 272] ---
        row_subtotal_counts = np.sum(
            self._base_counts[row_subtotal.addend_idxs, :], axis=0
        )
        intersection_count = np.sum(row_subtotal_counts[column_subtotal.addend_idxs])

        # --- `p` is intersection-table-proportion ---
        p = intersection_count / self._table_margin
        table_proportion_variance = p * (1 - p)

        return np.sqrt(table_proportion_variance / self._table_margin)

    def _subtotal_column(self, subtotal):
        """Return (n_rows,) ndarray of table-stderr `subtotal` value."""
        # --- Cannot calculate if there are any subtrahends ---
        if len(subtotal.subtrahend_idxs) > 0:
            return np.full(self._nrows, np.nan)

        # --- column of base subtotal counts is 1D, like [44 148 283 72] ---
        subtotal_counts = np.sum(
            self._base_counts[:, subtotal.addend_idxs],
            axis=1,
        )

        # --- `p` is subtotal-table-proportions ---
        p = subtotal_counts / self._table_margin
        table_proportion_variance = p * (1 - p)

        return np.sqrt(table_proportion_variance / self._table_margin)

    def _subtotal_row(self, subtotal):
        """Return (n_cols,) ndarray of table-stderr `subtotal` value."""
        # --- Cannot calculate if there are any subtrahends ---
        if len(subtotal.subtrahend_idxs) > 0:
            return np.full(self._ncols, np.nan)

        # --- row of base subtotal counts is 1D, like [435 392 260 162] ---
        subtotal_counts = np.sum(self._base_counts[subtotal.addend_idxs, :], axis=0)

        # --- `p` is subtotal-table-proportions ---
        p = subtotal_counts / self._table_margin
        table_proportion_variance = p * (1 - p)

        return np.sqrt(table_proportion_variance / self._table_margin)

    @lazyproperty
    def _table_margin(self):
        """Scalar or ndarray table-margin of cube-result matrix."""
        return self._cube_result_matrix.table_margin


class ZscoreSubtotals(_BaseSubtotals):
    """Computes subtotal values for the z-score measure.

    This is only operative for a CAT_X_CAT cube-result; an MR dimension causes all
    subtotals to be computed (elsewhere) as `np.nan`.
    """

    def __init__(
        self,
        weighted_counts,
        dimensions,
        diff_cols_nan=True,
        diff_rows_nan=True,
    ):
        super(ZscoreSubtotals, self).__init__(weighted_counts, dimensions)
        self._weighted_counts = weighted_counts
        self._diff_cols_nan = diff_cols_nan
        self._diff_rows_nan = diff_rows_nan

    @classmethod
    def blocks(
        cls, weighted_counts, dimensions, diff_cols_nan=True, diff_rows_nan=True
    ):
        """Return base, row and col insertion, and intersection matrices.

        These are in the form ready for assembly.

        Keyword arguments:
        `diff_cols_nan` -- Overrides subtotal differences in the columns direction eg
        for column bases (default False)
        `diff_rows_nan` -- Overrides subtotal differences in the rows direction eg for
        row bases (default False)
        """
        return cls(weighted_counts, dimensions, diff_cols_nan, diff_rows_nan)._blocks

    @classmethod
    def intersections(
        cls, weighted_counts, dimensions, diff_cols_nan=True, diff_rows_nan=True
    ):
        """Return (n_row_subtotals, n_col_subtotals) ndarray of intersection values.

        An intersection value arises where a row-subtotal crosses a column-subtotal.

        Keyword arguments:
        `diff_cols_nan` -- Overrides subtotal differences in the columns direction eg
        for column bases (default False)
        `diff_rows_nan` -- Overrides subtotal differences in the rows direction eg for
        row bases (default False)
        """
        return cls(
            weighted_counts, dimensions, diff_cols_nan, diff_rows_nan
        )._intersections

    @lazyproperty
    def _base_counts(self):
        """2D float64 nparray of weighted counts."""
        return self._weighted_counts.weighted_counts

    @lazyproperty
    def _blocks(self):
        """base, row and col insertion, and intersection matrices."""
        return [
            [self._weighted_counts.zscores, self._subtotal_columns],
            [self._subtotal_rows, self._intersections],
        ]

    @lazyproperty
    def _columns_margin(self):
        """2D np.float64 ndarray of weighted-N for each matrix column/cell."""
        return self._weighted_counts.columns_margin

    def _intersection(self, row_subtotal, column_subtotal):
        """Return value for intersection of `row_subtotal` and `column_subtotal`."""
        # --- Cannot calculate if there are any subtrahends ---
        col_has_subs = len(column_subtotal.subtrahend_idxs) > 0
        row_has_subs = len(row_subtotal.subtrahend_idxs) > 0

        if (
            # --- Intersections of subtotal differences are undefined ---
            (col_has_subs and row_has_subs)
            # --- Also need to respect diff_cols_nan/diff_rows_nan
            or (col_has_subs and self._diff_cols_nan)
            or (row_has_subs and self._diff_rows_nan)
        ):
            return np.nan

        row_subtotal_counts = np.sum(
            self._base_counts[row_subtotal.addend_idxs, :], axis=0
        )

        # --- the weighted-counts version of this intersection cell ---
        intersection_count = np.sum(row_subtotal_counts[column_subtotal.addend_idxs])

        # --- subtotal-margin is scalar, like 547 ---
        subtotal_margin = np.sum(row_subtotal_counts)

        # --- opposite-margin is scalar because no MR dimensions ---
        opposite_margin = np.sum(self._columns_margin[column_subtotal.addend_idxs])

        # --- table_margin is scalar because no MR dimensions ---
        table_margin = self._table_margin

        # --- expected_count is scalar ---
        expected_count = opposite_margin * subtotal_margin / table_margin

        # --- residuals is scalar ---
        residuals = intersection_count - expected_count

        # --- variance is scalar (I think) ---
        variance = (
            opposite_margin
            * subtotal_margin
            * ((table_margin - opposite_margin) * (table_margin - subtotal_margin))
            / table_margin ** 3
        )

        # --- result is scalar ---
        with np.errstate(divide="ignore", invalid="ignore"):
            return residuals / np.sqrt(variance)

    @lazyproperty
    def _ncols(self):
        """int count of columns in base-matrix."""
        return self._base_counts.shape[1]

    @lazyproperty
    def _nrows(self):
        """int count of rows in base-matrix."""
        return self._base_counts.shape[0]

    @lazyproperty
    def _rows_margin(self):
        """2D np.float64 ndarray of weighted-N for each matrix row/cell."""
        return self._weighted_counts.rows_margin

    def _subtotal_column(self, subtotal):
        """Return (n_rows,) ndarray of zscore `subtotal` value."""
        # --- Cannot calculate if there are any subtrahends ---
        if self._diff_cols_nan and len(subtotal.subtrahend_idxs) > 0:
            return np.full(self._nrows, np.nan)

        # --- the weighted-counts version of this subtotal-column ---
        subtotal_counts = np.sum(self._base_counts[:, subtotal.addend_idxs], axis=1)

        # --- subtotal-margin is scalar, like 547 ---
        subtotal_margin = np.sum(subtotal_counts)

        # --- base-rows-margin is 1D because no MR dimensions ---
        opposing_margin = self._rows_margin

        # --- table_margin is scalar because no MR dimensions ---
        table_margin = self._table_margin

        # --- expected_counts is 1D ---
        expected_counts = opposing_margin * subtotal_margin / table_margin

        # --- residuals is 1D, like: [ 11.04819 -37.72836  43.35049 -16.67031]
        residuals = subtotal_counts - expected_counts

        # --- variance is 1D, like: [12.413837 49.173948 53.980069 29.783739] ---
        variance = (
            opposing_margin
            * subtotal_margin
            * ((table_margin - opposing_margin) * (table_margin - subtotal_margin))
            / table_margin ** 3
        )

        # --- result is scalar or 1D, depending on dimensionality of residuals ---
        with np.errstate(divide="ignore", invalid="ignore"):
            return residuals / np.sqrt(variance)

    def _subtotal_row(self, subtotal):
        """Return (n_cols,) ndarray of z-score `subtotal` value."""
        # --- Cannot calculate if there are any subtrahends ---
        if self._diff_rows_nan and len(subtotal.subtrahend_idxs) > 0:
            return np.full(self._ncols, np.nan)

        # --- the weighted-counts version of this subtotal-row ---
        subtotal_counts = np.sum(self._base_counts[subtotal.addend_idxs, :], axis=0)

        # --- subtotal-margin is scalar, like 547 ---
        subtotal_margin = np.sum(subtotal_counts)

        # --- base-cols-margin is 1D because no MR dimensions ---
        opposing_margin = self._columns_margin

        # --- table_margin is scalar because no MR dimensions ---
        table_margin = self._table_margin

        # --- expected_counts is 1D ---
        expected_counts = opposing_margin * subtotal_margin / table_margin

        # --- residuals is 1D, like: [ 11.04819 -37.72836  43.35049 -16.67031]
        residuals = subtotal_counts - expected_counts

        # --- variance is 1D, like: [12.413837 49.173948 53.980069 29.783739] ---
        variance = (
            opposing_margin
            * subtotal_margin
            * ((table_margin - opposing_margin) * (table_margin - subtotal_margin))
            / table_margin ** 3
        )

        # --- result is scalar or 1D, depending on dimensionality of residuals ---
        with np.errstate(divide="ignore", invalid="ignore"):
            return residuals / np.sqrt(variance)

    @lazyproperty
    def _table_margin(self):
        """Scalar np.float64 weighted-N for overall table."""
        return self._weighted_counts.table_margin
